# -*- coding: utf-8 -*-
"""
TencentBlueKing is pleased to support the open source community by making
蓝鲸智云 - Resource SDK (BlueKing - Resource SDK) available.
Copyright (C) 2023 THL A29 Limited,
a Tencent company. All rights reserved.
Licensed under the MIT License (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at https://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
either express or implied. See the License for the
specific language governing permissions and limitations under the License.
We undertake not to change the open source license (MIT license) applicable
to the current version of the project delivered to anyone in the future.
"""

from bk_resource.base import Resource
from bk_resource.contrib import APIResource, BkApiResource, CacheResource
from bk_resource.management.root import adapter, api, resource

default_app_config = "bk_resource.apps.BKResourceConfig"

__author__ = "蓝鲸智云"
__copyright__ = "Copyright (c)   2012-2021 Tencent BlueKing. All Rights Reserved."
__doc__ = """
自动发现项目下resource和adapter和api
    cc
    ├── adapter
    │   ├── default.py
    │   ├── community
    │   │   └── resources.py
    │   └── enterprise
    │       └── resources.py
    └── resources.py
    使用:
        resource.cc -> cc/resources.py
        # 调用resource.cc 即可访问对应文件下的resource
        adapter.cc -> cc/adapter/default.py -> cc/adapter/${platform}/resources.py
        # 调用adapter.cc 既可访问对应文件下的resource，
        # 如果在${platform}/resources.py里面有相同定义，会重载default.py下的resource
    """
__all__ = [
    "Resource",
    "CacheResource",
    "APIResource",
    "BkApiResource",
    "adapter",
    "api",
    "resource",
    "contrib",
]



# -*- coding: utf-8 -*-
"""
TencentBlueKing is pleased to support the open source community by making
蓝鲸智云 - Resource SDK (BlueKing - Resource SDK) available.
Copyright (C) 2023 THL A29 Limited,
a Tencent company. All rights reserved.
Licensed under the MIT License (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at https://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
either express or implied. See the License for the
specific language governing permissions and limitations under the License.
We undertake not to change the open source license (MIT license) applicable
to the current version of the project delivered to anyone in the future.
"""

import inspect

from django.conf import settings
from django.template import engines

from bk_resource.management.root import ResourceShortcut
from bk_resource.utils.logger import logger

django_engine = engines["django"]


TEMPLATE = """# -*- coding: UTF-8 -*-

class ConfigureMixin(object):
    {% for module, methods in tree.items %}{% if module == "adapter" %}
    class adapter:
        {% for adapter_module, adapter_methods in methods.items %}
        class {{ adapter_module }}:{% for method_name, method in adapter_methods.items %}
            from {{ method.module }} import {{ method.name}} as {{ method_name }}{% endfor %}
            {% for method_name, method in adapter_methods.items %}{% if method.type == "function" %}
            {{ method_name }}: function = {{ method_name }}{% else %}
            {{ method_name }}: {{ method_name }} = ...{% endif %}{% endfor %}
            ...
        {% endfor %}
    {% else %}
    class {{ module }}:{% for method_name, method in methods.items %}
        from {{ method.module }} import {{ method.name}} as {{ method_name }}{% endfor %}
        {% for method_name, method in methods.items %}{% if method.type == "function" %}
        {{ method_name }}: function = {{ method_name }}{% else %}
        {{ method_name }}: {{ method_name }} = ...{% endif %}{% endfor %}
        ...
    {% endif %}{% endfor %}
    ...
"""

template = django_engine.from_string(TEMPLATE)


def render(define_tree, tab=0):
    result = ""
    # 初始化
    if tab == 0:
        result += "# -*- coding: UTF-8 -*-\n\nclass ConfigureMixin(object):\n"
        result += render(define_tree, tab + 1)
        result += "    ...\n"
        return result

    # 缩进数量计算
    tab_string = "    " * tab

    # 引入依赖
    for key, value in list(define_tree.items()):
        if value.get("type") in ["class", "function"]:
            result += "{}from {} import {} as {}\n".format(tab_string, value["module"], value["name"], key)

    # 函数及类定义
    for key, value in list(define_tree.items()):
        if value.get("type") == "class":
            result += "{}{}: {} = ...\n".format(tab_string, key, key)
        elif value.get("type") == "function":
            result += "{}{}: function = {}\n".format(tab_string, key, key)

    # 递归定义子类
    for key, value in list(define_tree.items()):
        if value.get("type") not in ["class", "function"]:
            result += "\n{}class {}:\n".format(tab_string, key)
            result += render(value, tab + 1)
            result += "%s    ...\n" % tab_string

    return result


def generate_stub_file():
    import os

    from bk_resource import api, resource

    def search_attr(instance):
        resource_tree = {}

        for name in dir(instance):
            if name.startswith("_"):
                continue
            obj = getattr(instance, name)

            if not isinstance(obj, ResourceShortcut):
                continue

            if hasattr(obj, "list_method"):
                try:
                    obj._setup()
                except Exception as err:
                    logger.exception(err)
                resource_tree[name] = {}
                for key, value in list(obj._methods.items()):
                    resource_tree[name][key] = {
                        "name": value.__name__,
                        "module": value.__module__,
                        "cls": value,
                    }
                    if inspect.isfunction(value):
                        resource_tree[name][key]["type"] = "function"
                    else:
                        resource_tree[name][key]["type"] = "class"
            else:
                resource_tree[name] = search_attr(obj)

        return resource_tree

    tree = search_attr(resource)
    tree.update(search_attr(api))

    file_dir = os.path.join(settings.BASE_DIR, "stdlib")
    file_path = os.path.join(file_dir, "resources.pyi")
    if not os.path.exists(file_dir):
        os.mkdir(file_dir)

    with open(file_path, "w+") as fp:
        fp.write(render(tree))



# -*- coding: utf-8 -*-
"""
TencentBlueKing is pleased to support the open source community by making
蓝鲸智云 - Resource SDK (BlueKing - Resource SDK) available.
Copyright (C) 2023 THL A29 Limited,
a Tencent company. All rights reserved.
Licensed under the MIT License (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at https://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
either express or implied. See the License for the
specific language governing permissions and limitations under the License.
We undertake not to change the open source license (MIT license) applicable
to the current version of the project delivered to anyone in the future.
"""

from blueapps.core.exceptions import BlueException
from django.utils.translation import gettext, gettext_lazy
from rest_framework import status


class CoreException(Exception):
    code = 100
    status_code = 500
    default_detail = gettext_lazy("服务异常")

    def __init__(self, message=None, data=None, code=None):
        """
        :param message: 错误信息
        :param data: 错误数据
        :param code: 错误码
        """
        if message is None:
            message = self.default_detail
        self.detail = message
        self.message = message

        if data is not None:
            self.data = data

        if code is not None:
            self.code = code

    def __str__(self):
        return str(self.message)


class APIRequestError(CoreException):
    code = 101
    message = gettext_lazy("API请求错误")

    def __init__(self, module_name: str = "", url: str = "", status_code: int = None, result=None):
        """
        :param result: 错误消息
        """
        if not isinstance(result, dict):
            result = {"message": result}
        if status_code:
            self.status_code = status_code
        result.update({"module_name": module_name, "url": url})

        self.data = result
        self.message = result.get("message")


class CustomError(CoreException):
    code = 102
    message = gettext_lazy("未知错误")


class DrfApiError(CoreException):
    code = 103
    message = gettext_lazy("REST API返回错误")

    @staticmethod
    def drf_error_processor(detail):
        """
        将DRF ValidationError 错误信息转换为字符串
        """
        if isinstance(detail, str):
            return detail
        elif isinstance(detail, dict):
            for k, v in list(detail.items()):
                if v:
                    return DrfApiError.drf_error_processor(v)
            else:
                return ""
        elif isinstance(detail, list):
            for item in detail:
                if item:
                    return DrfApiError.drf_error_processor(item)
            else:
                return ""
        else:
            return gettext("错误消息解析错误")


class HTTP404Error(CoreException):
    code = 104
    status_code = 404
    message = gettext_lazy("资源不存在")


class UserInfoMissing(CoreException):
    code = 105
    message = gettext_lazy("缺少用户信息")


class PermissionException(CoreException):
    code = 106
    status_code = 403
    default_detail = gettext_lazy("权限不足")


class ValidateException(CoreException):
    """校验失败"""

    code = 107
    default_detail = gettext_lazy("数据校验失败")


class PlatformAuthParamsNotExist(CoreException):
    code = 108
    default_detail = gettext_lazy("平台鉴权参数未配置")


class IAMNoPermission(BlueException):
    PLATFORM_CODE = "99"
    ERROR_CODE = "403"
    MESSAGE = gettext_lazy("权限不足")
    STATUS_CODE = status.HTTP_403_FORBIDDEN



# -*- coding: utf-8 -*-
"""
TencentBlueKing is pleased to support the open source community by making
蓝鲸智云 - Resource SDK (BlueKing - Resource SDK) available.
Copyright (C) 2023 THL A29 Limited,
a Tencent company. All rights reserved.
Licensed under the MIT License (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at https://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
either express or implied. See the License for the
specific language governing permissions and limitations under the License.
We undertake not to change the open source license (MIT license) applicable
to the current version of the project delivered to anyone in the future.
"""

import logging
from typing import Callable, Dict, Generic, Optional, TypeVar, cast

logger = logging.getLogger(__name__)


class Teapot:
    def __init__(self, factory: "GenericFactory"):
        self.__factory = factory

    def __getattr__(self, name: str):
        factory = self.__factory
        raise NotImplementedError(f"I'm a teapot from factory {factory.name}, I don't have {name}")


T_Type = TypeVar("T_Type")
T_Value = TypeVar("T_Value")
T_Instance = TypeVar("T_Instance")


class GenericFactory(Generic[T_Type, T_Value]):
    """
    通用泛型工厂实现

    >>> class MockType(Enum):
    ...     Registered = "A"
    ...     NotRegister = "B"

    >>> factory: Factory[MockType, str] = Factory("testing", "")
    """

    factory_type = "generic"

    def __init__(self, name: str, defaults: Optional[T_Value] = None, allow_overwrite: bool = False):
        self.values: Dict[Optional[T_Type], T_Value] = {}
        self.allow_overwrite = allow_overwrite
        self.name = name

        if defaults is None:
            self.defaults = cast(T_Value, Teapot(self))
        else:
            self.defaults = defaults

    def replace_defaults(self, defaults: T_Value) -> T_Value:
        """
        替换默认值，返回旧值
        """

        if not self.allow_overwrite:
            raise RuntimeError(f"default value of {self.factory_type} factory {self.name} already exists")

        logger.info("replace defaults of %s factory %s", self.factory_type, self.name)

        legacy_instance = self.defaults
        self.defaults = defaults
        return legacy_instance

    def register(self, typ: T_Type, value: T_Value):
        """
        注册类型

        :raises RuntimeError: 如果类型已经注册，且不允许覆盖
        """

        if typ in self.values and not self.allow_overwrite:
            raise RuntimeError(f"{typ} of {self.factory_type} factory {self.name} already exists")

        logger.info("register %s of %s factory %s", typ, self.factory_type, self.name)
        self.values[typ] = value

    def get(self, typ: Optional[T_Type] = None) -> T_Value:
        """获取指定类型的值，如果不存在则返回默认值"""

        return self.values.get(typ, self.defaults)

    def must_get(self, typ: T_Type) -> T_Value:
        """
        获取指定类型的值，如果不存在则抛出异常

        :raises RuntimeError: 如果未注册
        """

        if typ not in self.values:
            raise RuntimeError(f"{typ} of {self.factory_type} factory {self.name} not exists")

        return self.values[typ]

    def is_support(self, typ: T_Type) -> bool:
        """是否支持该类型"""

        return typ in self.values

    def clear(self):
        """清空注册的类型"""

        logger.warning("%s factory %s cleared", self.factory_type, self.name)
        self.values.clear()

    def __getitem__(self, key: T_Type) -> T_Value:
        """按下标获取类型的值，相当于 self.must_get(key)"""
        return self.must_get(key)


class SingletonFactory(Generic[T_Type, T_Instance], GenericFactory[T_Type, T_Instance]):
    """实例工厂类，用于注册已经初始化的实例对象"""

    factory_type = "singleton"

    def __call__(self, typ: Optional[T_Type] = None) -> T_Instance:
        """获取指定类型的实例，相当于 self.get(typ)"""
        return self.get(typ)


class SimpleFactory(Generic[T_Type, T_Instance], GenericFactory[T_Type, Callable[..., T_Instance]]):
    """简单工厂类，用于注册创建实例的回调函数"""

    factory_type = "simple"

    def make(self, typ: T_Type, *args, **kwargs) -> T_Instance:
        """构造对应类型的实例"""

        callback = self.get(typ)

        return callback(*args, **kwargs)

    def must_make(self, typ: T_Type, *args, **kwargs) -> T_Instance:
        """构造对应类型的实例，如果不存在则抛出异常"""

        callback = self.must_get(typ)

        return callback(*args, **kwargs)

    def __call__(self, typ: Optional[T_Type] = None, *args, **kwargs) -> T_Instance:
        """构造对应类型的实例"""

        callback = self.get(typ)

        return callback(*args, **kwargs)



# -*- coding: utf-8 -*-
"""
TencentBlueKing is pleased to support the open source community by making
蓝鲸智云 - Resource SDK (BlueKing - Resource SDK) available.
Copyright (C) 2023 THL A29 Limited,
a Tencent company. All rights reserved.
Licensed under the MIT License (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at https://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
either express or implied. See the License for the
specific language governing permissions and limitations under the License.
We undertake not to change the open source license (MIT license) applicable
to the current version of the project delivered to anyone in the future.
"""

import datetime
import re
import time
from typing import Union

import arrow
from django.conf import settings
from django.utils import timezone


def now():
    return timezone.now()


def now_str(_format="%Y-%m-%d %H:%M:%S"):
    return timezone.now().strftime(_format)


def localtime(value):
    """
    to local time
    :param value: datetime obj
    :return: 返回带本地(业务)时区的datetime对象
    """
    if timezone.is_aware(value):
        return timezone.localtime(value)
    return timezone.make_aware(value)


def utc2localtime(utc_time):
    """
    to local time
    :param utc_time: 时间戳
    :return: 返回带本地(业务)时区的datetime对象
    """
    datetime_value = datetime.datetime.fromtimestamp(utc_time)
    return localtime(datetime_value)


def timestamp2datetime(timestamp):
    """
    to datetime obj
    :param timestamp: 时间戳
    :return: 返回datetime对象
    """
    return datetime.datetime.fromtimestamp(timestamp)


def mysql_time(value):
    """
    to mysql time
    :param value: datetime obj
    :return: 返回带本地(业务)时区的datetime对象
    """
    if settings.USE_TZ:
        if timezone.is_aware(value):
            return value
        else:
            return arrow.get(value).replace(tzinfo=timezone.get_current_timezone_name()).datetime
    else:
        if timezone.is_aware(value):
            return arrow.get(value).to(timezone.get_current_timezone_name()).naive
        else:
            return value


def biz_time_zone_offset():
    """
    获取业务的时区偏移
    """
    offset = arrow.now().replace(tzinfo=timezone.get_current_timezone().zone).format("Z")
    return str(offset[0]) + str(int(offset[1:3]))  # 转成小时的精度, 而不是分钟


def strftime_local(value, _format="%Y-%m-%d %H:%M:%S%z"):
    """
    转成业务时区字符串
    :param value: datetime obj
    """
    return localtime(value).strftime(_format)


def biz2utc_str(local_time, _format="%Y-%m-%d %H:%M:%S"):
    """
    功能: 业务时间字符串 转换成 零时区字符串
    场景: 界面传过来的时间需要转换成零时区去查询db或调用API
    """
    return arrow.get(local_time).replace(tzinfo=timezone.get_current_timezone().zone).to("utc").strftime(_format)


def utc2biz_str(utc_time, _format="%Y-%m-%d %H:%M:%S"):
    """
    功能: 零时区字符串 转换成 业务时间字符串
    场景: 从DB或调用API传回来的时间(utc时间)需要转换成业务时间再给到前端显示
    """
    return arrow.get(utc_time).to(timezone.get_current_timezone().zone).strftime(_format)


def utc2_str(utc_time, _format="%Y-%m-%d %H:%M:%S"):
    """
    将utc时间戳转换为utc时间格式
    :param utc_time: 时间戳
    :param _format:转换格式
    :return:
    """
    return arrow.get(utc_time).strftime("%Y-%m-%d %H:%M:%S")


def get_timestamp_range_by_biz_date(date):
    """
    功能: 解析从浏览器传过来的日期格式字符串, 转成时间戳timestamp格式
    @return tuple(timestamp(unit: s), timestamp(unit: s))
    """
    start_timestamp = arrow.get(date).replace(tzinfo=timezone.get_current_timezone().zone).timestamp()
    end_timestamp = start_timestamp + 86400  # 24 * 3600
    return start_timestamp, end_timestamp


def gen_default_time_range(days=1, offset=0):
    """
    功能: 生成默认的业务时间日期范围
    """
    biz_today = timezone.localtime(timezone.now()).date()
    if days + offset < 1:
        return biz_today, biz_today + datetime.timedelta(1)
    return biz_today - datetime.timedelta(days + offset - 1), biz_today + datetime.timedelta(1 - offset)


def parse_time_range(time_range=None, days=1, offset=0, _time_format="%Y-%m-%d %H:%M"):
    """
    功能: 解析从浏览器传过来的time_range, 转成UTC时区, 再返回timestamp
    场景: 获取图表数据
    @return tuple(timestamp(unit: s), timestamp(unit: s))
    """
    if not time_range:
        # 默认取业务时区当天的数据
        start, end = gen_default_time_range(days, offset)
    else:
        time_range = time_range.replace("&nbsp;", " ")
        start, end = [i.strip() for i in time_range.split("--")]
        # start = date_convert(start, 'datetime', _time_format)
        # end = date_convert(end, 'datetime', _time_format)
    tzinfo = timezone.get_current_timezone().zone
    start = arrow.get(start)
    end = arrow.get(end)

    # 没有时区信息时，使用业务时区补全
    if not hasattr(start.tzinfo, "_offset"):
        start = start.replace(tzinfo=tzinfo)
    if not hasattr(end.tzinfo, "_offset"):
        end = end.replace(tzinfo=tzinfo)
    return start.timestamp, end.timestamp


def date_convert(_date, _format, _time_format="%Y-%m-%d %H:%M:%S"):
    try:
        if isinstance(_date, datetime.date):
            if _format == "utc":
                return datetime2utc(_date, _time_format)

        elif isinstance(_date, str):
            if _format == "datetime":
                return str2datetime(_date, _time_format)
            elif _format == "date":
                return str2date(_date, "%Y-%m-%d")
            elif _format == "utc":
                return str2utc(_date, _time_format)

        elif isinstance(_date, int):
            if _format == "datetime":
                return utc2datetime(_date, _time_format)
            elif _format == "date":
                return utc2date(_date, _time_format)

        return _date
    except Exception:
        return ""


# 字符串转datetime
def str2datetime(_str, _format="%Y-%m-%d %H:%M:%S"):
    return datetime.datetime.strptime(_str, _format)


def str2date(_str, _format="%Y-%m-%d %H:%M:%S"):
    return datetime.datetime.strptime(_str, _format).date()


def str2utc(_str, _format="%Y-%m-%d %H:%M:%S"):
    """
    字符串转UTC
    """
    _datetime = datetime.datetime.strptime(_str, _format)
    return int(time.mktime(_datetime.timetuple()))


def datetime2utc(_datetime, _format="%Y-%m-%d %H:%M:%S"):
    """
    datetime转UTC
    """
    return int(time.mktime(_datetime.timetuple()))


def date2str(_datetime, _format="%Y-%m-%d") -> str:
    """
    传入一个标准日期，返回其字符串形式
    """
    return _datetime.strftime(_format)


def utc2datetime(_utc, _format="%Y-%m-%d %H:%M:%S"):
    """
    UTC转datetime
    """
    return time.strftime(_format, time.localtime(int(_utc)))


def utc2date(_utc, _format="%Y-%m-%d %H:%M"):
    """
    UTC转date
    """
    return time.strftime(_format, time.localtime(int(_utc)))


def get_datetime_range(period, distance, now=None, rounding=True):
    now = now or localtime(timezone.now())
    if period == "day":
        begin = arrow.get(now) - datetime.timedelta(days=distance)
        end = arrow.get(now)
        if rounding:
            begin = begin.ceil("day")
            end = end.ceil("day")
    elif period == "hour":
        begin = arrow.get(now) - datetime.timedelta(hours=distance)
        end = arrow.get(now)
        if rounding:
            begin = begin.ceil("hour")
            end = end.ceil("hour")
    else:
        raise TypeError("invalid period: %r" % period)

    if settings.USE_TZ:
        return begin.datetime, end.datetime
    else:
        return begin.naive, end.naive


def get_datetime_list(begin_time, end_time, period):
    if period == "day":
        timedelta = "days"
    elif period == "hour":
        timedelta = "hours"
    else:
        raise TypeError("invalid period: %r" % period)

    datetime_list = []
    item = begin_time
    while item < end_time:
        datetime_list.append(item)
        item = item + timezone.timedelta(**{timedelta: 1})

    return datetime_list


def utcoffset_in_seconds():
    """
    获取当前时区偏移量（秒）
    """
    return localtime(now()).utcoffset().total_seconds()


def datetime2timestamp(value):
    """
    datetime转timestamp格式
    """
    return time.mktime(localtime(value).timetuple())


def hms_string(
    sec_elapsed,
    display_num=2,
    day_unit="d",
    hour_unit="h",
    minute_unit="m",
    second_unit="s",
):
    """
    将秒数转化为 人类易读时间 1d 12h 3m
    """
    d = int(sec_elapsed // (60 * 60) // 24)
    h = int((sec_elapsed - d * 24 * 3600) // 3600)
    m = int((sec_elapsed - d * 24 * 3600 - h * 3600) // 60)
    s = int(sec_elapsed - d * 24 * 3600 - h * 3600 - m * 60)

    units = [(d, day_unit), (h, hour_unit), (m, minute_unit), (s, second_unit)]
    time_str = ""
    index = 0
    for unit in units:
        if unit[0] == 0:
            continue

        index += 1
        time_str += "{}{} ".format(unit[0], unit[1])
        if index >= display_num:
            return time_str.strip()

    if not time_str:
        return "{}{}".format(s, second_unit)

    return time_str.strip()


TIME_ABBREVIATION_MATCH = re.compile(r"^[-+]?\d*\.?\d+[smhdwMy]")


def parse_time_compare_abbreviation(time_offset: Union[int, str]) -> int:
    """
    时间对比解析(h/d/w/M/y)
    :param time_offset: 1d / 0.5m / -5w
    :return: 返回一个整数，以秒为单位。 int (unit: s)
    """
    if isinstance(time_offset, (int, float)):
        return time_offset

    time_offset = str(time_offset).strip()

    if not time_offset or not TIME_ABBREVIATION_MATCH.match(time_offset):
        time_offset = 0
    else:
        if time_offset[-1] in "smhdw":
            time_unit = {"s": 1, "m": 60, "h": 3600, "d": 86400, "w": 604800}
            time_offset = -int(float(time_offset[:-1]) * time_unit[time_offset[-1]])
        else:
            current_time = arrow.now()
            if time_offset[-1] == "M":
                last_time = current_time.replace(month=-int(float(time_offset[:-1])))
            else:
                last_time = current_time.replace(year=-int(float(time_offset[:-1])))
            time_offset = int((last_time - current_time).total_seconds())

    return time_offset


def datetime_str_to_datetime(datetime_str, format, time_zone=0):
    return arrow.get(datetime.datetime.strptime(datetime_str, format)).replace(hour=time_zone).datetime



# -*- coding: utf-8 -*-
"""
TencentBlueKing is pleased to support the open source community by making
蓝鲸智云 - Resource SDK (BlueKing - Resource SDK) available.
Copyright (C) 2023 THL A29 Limited,
a Tencent company. All rights reserved.
Licensed under the MIT License (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at https://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
either express or implied. See the License for the
specific language governing permissions and limitations under the License.
We undertake not to change the open source license (MIT license) applicable
to the current version of the project delivered to anyone in the future.
"""

import base64
import datetime
import decimal
import hashlib
import json
import os
import pkgutil
import re
import socket
import sys
import traceback
import uuid
from collections import OrderedDict, defaultdict
from contextlib import contextmanager
from io import StringIO
from pipes import quote
from zipfile import ZipFile

from django.utils.encoding import force_str
from django.utils.functional import Promise
from django.utils.timezone import is_aware

from bk_resource.utils import time_tools
from bk_resource.utils.logger import logger

IDS_REGEX = re.compile(r"^\d+(,\d+)*$")


def package_contents(package):
    if isinstance(package, str):
        return package_contents(__import__(package, fromlist=[str("*")]))
    return [name for _, name, _ in pkgutil.iter_modules([os.path.dirname(package.__file__)])]


class DictObj(object):
    __non_zero = False

    def __init__(self, kwargs=None):
        if kwargs is None:
            kwargs = dict()
        self.__dict__.update(kwargs)
        for k, v in kwargs.items():
            if not self.__non_zero:
                self.__non_zero = True
            try:
                setattr(self, k, v)
            except AttributeError:
                msg = "[%s] attribute: `%s` has already exists, " "check your class definition `@property`" % (
                    self.__class__.__name__,
                    k,
                )
                raise AttributeError(msg)

    def __str__(self):
        return json.dumps(self.__dict__, ensure_ascii=False)

    def __getattr__(self, item):
        return self.__dict__.get(item, None)

    def __bool__(self):
        return self.__non_zero


class DatetimeEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, datetime.datetime):
            return time_tools.strftime_local(obj)
        elif isinstance(obj, datetime.date):
            return obj.strftime("%Y-%m-%d")
        elif isinstance(obj, datetime.time):
            if is_aware(obj):
                raise ValueError("JSON can't represent timezone-aware times.")
            r = obj.isoformat()
            if obj.microsecond:
                r = r[:12]
            return r
        if isinstance(obj, decimal.Decimal):
            return str(obj)
        if isinstance(obj, set):
            return list(obj)
        if isinstance(obj, Promise):
            return force_str(obj)
        if issubclass(obj.__class__, DictObj):
            return obj.__dict__
        if isinstance(obj, bytes):
            return obj.decode("utf-8")
        return json.JSONEncoder.default(self, obj)


@contextmanager
def ignored(*exceptions, **kwargs):
    try:
        yield
    except exceptions:
        if kwargs.get("log_exception", True):
            logger.warning(traceback.format_exc())
        pass


def ok(message="", **options):
    result = {"result": True, "message": message, "msg": message}
    result.update(**options)
    return result


def failed(message="", **options):
    if not isinstance(message, str):
        if isinstance(message, str):
            message = message.encode()
        message = str(message)
    result = {"result": False, "message": message, "data": {}, "msg": message}
    result.update(**options)
    return result


def failed_data(message, data, **options):
    if not isinstance(message, str):
        if isinstance(message, str):
            message = message.encode()
        message = str(message)
    result = {"result": False, "message": message, "data": data, "msg": message}
    result.update(**options)
    return result


def ok_data(data=None, **options):
    if data is None:
        data = {}
    result = {"result": True, "message": "", "data": data, "msg": ""}
    result.update(**options)
    return result


def get_unique_list(_list):
    """
    list去重，并保持原有数据顺序
    :param _list:
    :return:
    """
    return list(OrderedDict.fromkeys(_list))


def today_start_timestamp(the_day=None):
    if the_day is None:
        the_day = datetime.date.today()
    if isinstance(the_day, datetime.datetime):
        the_day = the_day.date()
    days = (the_day - datetime.date(1970, 1, 1)).days
    return days * 3600 * 24 * 1000


def dict_slice(adict, start, end):
    """
    字典切片
    :param adict:
    :param start:
    :param end:
    """
    keys = list(adict.keys())
    dict_slice = {}
    for k in keys[start:end]:
        dict_slice[k] = adict[k]
    return dict_slice


def get_first(objs, default=""):
    """get the first element in a list or get blank"""
    if len(objs) > 0:
        return objs[0]
    return default


def get_list(obj):
    return obj if isinstance(obj, list) else [obj]


def get_one(obj):
    return obj[0] if isinstance(obj, (list, tuple)) else obj


def uniqid():
    # 不带横杠
    return uuid.uuid3(uuid.uuid1(), uuid.uuid4().hex).hex


def uniqid4():
    # 带横杠
    return str(uuid.uuid4())


def file_read(filename):
    """
    打开utf-8编码文件
    """
    with open(filename) as f:
        return f.read()


def file_md5sum(file):
    """
    计算文件的md5值
    """
    # 文件指针指向开头
    file.seek(0, 0)

    md5 = hashlib.md5()
    for chunk in file.chunks():
        md5.update(chunk)

    file.seek(0, 0)
    return md5.hexdigest()


def file_rename(file, new_file_name=None):
    """
    文件重命名，保留后缀
    :param file: 文件
    :param new_file_name: 新文件名，没有则默认取文件的md5
    :return: 新文件名
    """
    new_file_name = new_file_name or file_md5sum(file)
    ext = ""
    if "." in file.name:
        ext = file.name.split(".")[-1]

    if ext:
        new_file_name = "{}.{}".format(new_file_name, ext)
    return new_file_name


def tree():
    return defaultdict(tree)


def _count_md5(content):
    if content is None:
        return None
    m2 = hashlib.md5()
    if isinstance(content, str):
        m2.update(content.encode("utf8"))
    else:
        m2.update(content)
    return m2.hexdigest()


def count_md5(content, dict_sort=True):
    if dict_sort and isinstance(content, dict):
        # dict的顺序受到hash的影响，所以这里先排序再计算MD5
        return count_md5([(str(k), count_md5(content[k])) for k in sorted(content.keys())])
    elif isinstance(content, (list, tuple)):
        content = sorted([count_md5(k) for k in content])
    return _count_md5(str(content))


def get_md5(content):
    if isinstance(content, list):
        return [count_md5(c) for c in content]
    else:
        return count_md5(content)


REG_SPLIT_LIST = re.compile(r"\s*[;,]\s*")


def split_list(raw_string):
    if isinstance(raw_string, (tuple, list, set)):
        return raw_string
    return [x for x in REG_SPLIT_LIST.split(raw_string) if x]


def get_local_ip():
    """
    Returns the actual ip of the local machine.
    This code figures out what source address would be used if some traffic
    were to be sent out to some well known address on the Internet. In this
    case, a Google DNS server is used, but the specific address does not
    matter much.  No traffic is actually sent.

    stackoverflow上有人说用socket.gethostbyname(socket.getfqdn())
    但实测后发现有些机器会返回127.0.0.1
    """
    try:
        csock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        csock.connect(("8.8.8.8", 80))
        (addr, port) = csock.getsockname()
        csock.close()
        return addr
    except socket.error:
        return "127.0.0.1"


def number_format(v):
    try:
        # 字符型转为数值型 其他保持不变
        if isinstance(v, str):
            if v.find(".") > -1:
                try:
                    return float(v)
                except:  # noqa
                    return v
            else:
                try:
                    return int(v)
                except:  # noqa
                    return v
        elif v:
            return v
        else:
            return 0
    except Exception as e:
        raise Exception(e)


def convert_to_cmdline_args_str(kv_dict):
    """
    命令行参数组装
    :param kv_dict: 参数key-value对
    {
        "version": None,
        "-s": "xxx",
        "--some": "yyy",
    }
    :return:
    """
    result = ""
    for k, v in list(kv_dict.items()):
        if v:
            # 进行shell转义
            v = quote(v)
        else:
            v = ""
        if k.startswith("--"):
            result += "{}={} ".format(k, v)
        else:
            result += "{} {} ".format(k, v)
    return result


def escape_cmd_argument(arg):
    # Escape the argument for the cmd.exe shell.
    # First we escape the quote chars to produce a argument suitable for
    # CommandLineToArgvW. We don't need to do this for simple arguments.

    # if not arg or re.search(r'(["\s])', arg):
    #     arg = '"' + arg.replace('"', r'\"') + '"'

    meta_chars = '()%!^"<>&|'
    meta_re = re.compile("(" + "|".join(re.escape(char) for char in list(meta_chars)) + ")")
    meta_map = {char: "^%s" % char for char in meta_chars}

    def escape_meta_chars(m):
        char = m.group(1)
        return meta_map[char]

    return meta_re.sub(escape_meta_chars, arg)


def extract_zip(input_zip):
    """
    解压文件，获取文件列表
    """
    input_zip = ZipFile(input_zip)
    return {name: StringIO(input_zip.read(name)) for name in input_zip.namelist()}


def convert_img_to_base64(image, format="PNG"):
    """
    :param image: Image图片对象
    :param format: 保存格式
    :return: base64 string
    """
    img_buffer = StringIO()
    image.save(img_buffer, format=format, quality=95)
    base64_value = base64.b64encode(img_buffer.getvalue())
    return "data:image/{format};base64,{value}".format(format=format.lower(), value=base64_value)


def parse_filter_condition_dict(settings_condition_item, filter_key):
    if "method" in settings_condition_item:
        filter_key = "{}__{}".format(filter_key, settings_condition_item["method"])
    if "sql_statement" not in settings_condition_item:
        return None, None
    return filter_key, settings_condition_item["sql_statement"]


def safe_int(int_str, dft=0):
    try:
        int_val = int(int_str)
    except Exception:
        try:
            int_val = int(float(int_str))
        except Exception:
            int_val = dft
    return int_val


def safe_float(value):
    try:
        return float(value)
    except (ValueError, TypeError):
        return float("nan")


def proxy(obj):
    class Proxy(object):
        def __getattribute__(self, item):
            return getattr(obj, item)

    return Proxy()


# create a new context for this task
ctx = decimal.Context()

# 20 digits should be enough for everyone :D
ctx.prec = 20


def float_to_str(f):
    """
    Convert the given float to a string,
    without resorting to scientific notation
    """
    d1 = ctx.create_decimal(repr(f))
    return format(d1, "f")


def to_dict(obj):
    """
    python 对象递归转成字典
    """
    if isinstance(obj, dict):
        data = {}
        for (k, v) in list(obj.items()):
            data[k] = to_dict(v)
        return data
    elif hasattr(obj, "__iter__") and not isinstance(obj, str):
        return [to_dict(v) for v in obj]
    elif hasattr(obj, "__dict__"):
        data = {}
        for key in dir(obj):
            value = getattr(obj, key)
            if not key.startswith("_") and not callable(value):
                data[key] = to_dict(value)
        return data
    else:
        return obj


def replce_special_val(s, replace_dict):
    """
    替换特殊变量
    :param s: 待替换字符串
    :param replace_dict: 替换映射
    :return: 替换结果
    """
    for key, value in replace_dict.items():
        s = s.replace(key, value)
    return s


def is_backend():
    basename = os.path.basename(sys.argv[0])

    # 非web请求
    if any(
        [
            "manage.py" == basename and "runserver" not in sys.argv,
            "celery" in sys.argv,
            "test" in sys.argv,
            "migrate" in sys.argv,
            basename in ["django_test_manage.py", "pydevconsole.py"],
            basename.find("pytest") != -1,
        ]
    ):
        return True

    return False



# -*- coding: utf-8 -*-
"""
TencentBlueKing is pleased to support the open source community by making
蓝鲸智云 - Resource SDK (BlueKing - Resource SDK) available.
Copyright (C) 2023 THL A29 Limited,
a Tencent company. All rights reserved.
Licensed under the MIT License (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at https://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
either express or implied. See the License for the
specific language governing permissions and limitations under the License.
We undertake not to change the open source license (MIT license) applicable
to the current version of the project delivered to anyone in the future.
"""

import abc
from typing import Union

from django.core.handlers.wsgi import WSGIRequest
from django.db import models
from django.utils.translation import gettext
from rest_framework.generics import GenericAPIView
from rest_framework.viewsets import ModelViewSet

from bk_resource import Resource
from bk_resource.utils.request import get_moke_request


class Empty:
    ...


class ViewMixin(GenericAPIView):
    filter_fields = []
    search_fields = []
    view_set_attrs = {}

    @property
    def model(self):
        """
        关联的Django Model
        """
        raise NotImplementedError

    @property
    def queryset(self):
        """
        关联的QuerySet
        """
        if hasattr(self.model, "is_deleted"):
            return self.model.objects.filter(is_deleted=False)
        return self.model.objects.all()

    @property
    def serializer_class(self):
        """
        序列化器
        """
        raise NotImplementedError

    def build_request(self, *, method: str, params: dict) -> WSGIRequest:
        # 构造虚拟请求对象
        request = get_moke_request(REQUEST_METHOD=method)
        # 为对应的方法添加请求体
        if request.method.lower() in ["get"]:
            setattr(request, "query_params", params)
        elif request.method.lower() in ["post", "put", "patch"]:
            setattr(request, "data", params)
        return request

    def build_view_set(self, *, request: WSGIRequest = None, method: str = None, params: dict) -> ModelViewSet:
        # 构造 Request
        if request is None:
            assert method is not None, gettext("request and method cannot be empty at the same time")
            request = self.build_request(method=method, params=params)
        # 构造 ViewSet, ModelResource分页配置与ResourceViewSet冲突
        model_params = {
            "queryset": self.queryset,
            "filter_backends": self.filter_backends,
            "serializer_class": self.serializer_class,
            "pagination_class": None,
            "lookup_field": self.lookup_field,
            "kwargs": params,
            "format_kwarg": self.get_format_suffix(**params),
            "filter_fields": self.filter_fields,
            "search_fields": self.search_fields,
            "request": request,
            **self.view_set_attrs,
        }
        view_set = ModelViewSet(**model_params)
        return view_set

    def build_request_and_view_set(self, *, method: str, params: dict) -> (WSGIRequest, ModelViewSet):
        request = self.build_request(method=method, params=params)
        view_set = self.build_view_set(request=request, method=method, params=params)
        return request, view_set


class ModelResource(Resource, ViewMixin, metaclass=abc.ABCMeta):
    """
    ModelViewSet
    """

    @property
    def action(self):
        raise NotImplementedError(gettext("action not set"))

    def perform_request(self, validated_request_data: dict) -> any:
        try:
            handler = getattr(self, self.action)
        except (TypeError, AttributeError):
            raise NotImplementedError(gettext("handler of %s is not implemented") % self.action)
        return handler(validated_request_data)

    def list(self, params: dict) -> list:
        request, view_set = self.build_request_and_view_set(method="GET", params=params)
        return view_set.list(request, params).data

    def retrieve(self, params: dict) -> dict:
        request, view_set = self.build_request_and_view_set(method="GET", params=params)
        return view_set.retrieve(request, params).data

    def create(self, params: dict) -> dict:
        request, view_set = self.build_request_and_view_set(method="POST", params=params)
        return view_set.create(request, params).data

    def update(self, params: dict) -> dict:
        partial = params.pop("partial", False)
        request, view_set = self.build_request_and_view_set(method="PATCH" if partial else "PUT", params=params)
        return view_set.update(request, params, partial=partial).data

    def destroy(self, params: dict) -> None:
        request, view_set = self.build_request_and_view_set(method="DELETE", params=params)
        return view_set.destroy(request, params).data

    def get_val_from_request_data(self, request_data: Union[models.Model, dict], key: str) -> any:
        try:
            if isinstance(request_data, models.Model):
                return getattr(request_data, key)
            else:
                return request_data[key]
        except (AttributeError, KeyError):
            return Empty

    def build_extra_params(self, request_data: Union[models.Model, dict], validated_request_data: dict) -> dict:
        non_use_fields = set(self.filter_fields) - set(validated_request_data.keys())
        for field in non_use_fields:
            val = self.get_val_from_request_data(request_data, field)
            if val is Empty:
                continue
            validated_request_data[field] = val
        return validated_request_data



# -*- coding: utf-8 -*-
"""
TencentBlueKing is pleased to support the open source community by making
蓝鲸智云 - Resource SDK (BlueKing - Resource SDK) available.
Copyright (C) 2023 THL A29 Limited,
a Tencent company. All rights reserved.
Licensed under the MIT License (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at https://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
either express or implied. See the License for the
specific language governing permissions and limitations under the License.
We undertake not to change the open source license (MIT license) applicable
to the current version of the project delivered to anyone in the future.
"""

import abc
from typing import Dict

import requests
from django.utils.encoding import force_str
from django.utils.translation import gettext
from requests.exceptions import HTTPError

from bk_resource.contrib.cache import CacheResource
from bk_resource.exceptions import APIRequestError
from bk_resource.settings import bk_resource_settings
from bk_resource.utils.logger import logger


class ApiResourceProtocol(metaclass=abc.ABCMeta):
    """
    API Resource Protocol
    """

    @property
    @abc.abstractmethod
    def module_name(self):
        """
        模块名
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def base_url(self):
        """
        api 基本url生成规则
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def action(self):
        """
        url的后缀，通常是指定特定资源
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def method(self):
        """
        请求方法，仅支持GET或POST
        """
        raise NotImplementedError

    def build_url(self, validated_request_data: dict) -> str:
        raise NotImplementedError

    def build_request_data(self, validated_request_data: dict) -> dict:
        raise NotImplementedError

    def build_header(self, validated_request_data: dict) -> Dict[str, str]:
        raise NotImplementedError

    def before_request(self, context: dict) -> dict:
        raise NotImplementedError

    def parse_response(self, result: requests.Response) -> dict:
        raise NotImplementedError


class APIResource(ApiResourceProtocol, CacheResource, metaclass=abc.ABCMeta):
    """
    API类型的Resource
    """

    module_name = "default"
    TIMEOUT = 60
    IS_STANDARD_FORMAT = True
    url_keys = []

    def __init__(self, **kwargs):
        super(APIResource, self).__init__(**kwargs)
        assert self.method.upper() in ["GET", "POST", "PUT", "PATCH", "DELETE"], gettext(
            "%s method 仅支持GET或POST或PUT或PATCH或DELETE，当前为%s"
        ) % (self.module_name, self.method.upper())
        self.method = self.method.upper()
        self.session = requests.session()

    def request(self, request_data=None, **kwargs):
        request_data = request_data or kwargs
        return super(APIResource, self).request(request_data, **kwargs)

    def perform_request(self, validated_request_data):
        """
        发起http请求
        """
        validated_request_data = dict(validated_request_data)
        validated_request_data = self.build_request_data(validated_request_data)

        # 拼接最终请求的url
        request_url = self.build_url(validated_request_data)
        logger.debug("request: {}".format(request_url))

        try:
            headers = self.build_header(validated_request_data)
            kwargs = {
                "method": self.method,
                "url": request_url,
                "timeout": self.TIMEOUT,
                "headers": headers,
                "verify": bk_resource_settings.REQUEST_VERIFY,
            }
            if self.method == "GET":
                kwargs["params"] = validated_request_data
                kwargs = self.before_request(kwargs)
                request_url = kwargs.pop("url")
                if "method" in kwargs:
                    del kwargs["method"]
                response = self.session.get(request_url, **kwargs)
            else:
                non_file_data, file_data = self.split_request_data(validated_request_data)
                if not file_data:
                    # 不存在文件数据，则按照json方式去请求
                    kwargs["json"] = non_file_data
                else:
                    # 若存在文件数据，则将非文件数据和文件数据分开传参
                    kwargs["files"] = file_data
                    kwargs["data"] = non_file_data

                kwargs = self.before_request(kwargs)
                response = self.session.request(**kwargs)
        except Exception as err:
            logger.exception(f"APIRequestFailed => {err}")
            err_message = err.__doc__ or err.__class__.__name__
            raise APIRequestError(
                module_name=self.module_name,
                url=self.action,
                result=err_message,
            )
        return self.parse_response(response)

    def build_url(self, validated_request_data):
        """
        最终请求的url，可以由子类进行重写
        """
        url = self.base_url.rstrip("/") + "/" + self.action.lstrip("/")
        if self.url_keys:
            kvs = {_k: validated_request_data[_k] for _k in self.url_keys}
            url = url.format(**kvs)
        return url

    def build_request_data(self, validated_request_data):
        return validated_request_data

    def build_header(self, validated_request_data):
        """
        请求头，可以由子类重写
        """
        return {}

    def before_request(self, kwargs):
        return kwargs

    def parse_response(self, response: requests.Response):
        """
        在提供数据给response_serializer之前，对数据作最后的处理，子类可进行重写
        """
        try:
            result_json = response.json()
        except Exception as err:
            logger.exception("{} => {}".format(gettext("Response Parse Error"), err))
            result_json = response.content

        try:
            response.raise_for_status()
        except HTTPError as err:
            logger.exception(gettext("【%s】请求API错误：%s，url: %s ") % (self.module_name, err, response.request.url))
            content = str(err.response.content)
            if isinstance(result_json, dict):
                content = "[{code}] {message}".format(code=result_json.get("code"), message=result_json.get("message"))
            raise APIRequestError(
                module_name=self.module_name,
                url=self.action,
                status_code=response.status_code,
                result=content,
            )

        if not self.IS_STANDARD_FORMAT:
            return result_json

        if not isinstance(result_json, dict):
            raise APIRequestError(
                module_name=self.module_name,
                url=self.action,
                result=gettext("返回格式有误 => %s") % force_str(result_json),
            )

        if not result_json.get("result", True) and result_json.get("code") != 0:
            msg = result_json.get("message", "")
            request_id = result_json.pop("request_id", "") or response.headers.get("x-bkapi-request-id", "")
            logger.error(
                "【Module: %s】【Action: %s】(%s) get error：%s",
                self.module_name,
                self.action,
                request_id,
                msg,
                extra=dict(module_name=self.module_name, url=response.request.url),
            )
            raise APIRequestError(
                module_name=self.module_name,
                url=self.action,
                result=result_json,
            )
        return result_json.get("data")

    @staticmethod
    def split_request_data(data):
        """
        切分请求参数为文件/非文件类型，便于requests参数组装
        """
        file_data = {}
        non_file_data = {}
        for request_param, param_value in list(data.items()):
            if hasattr(param_value, "read"):
                # 一般认为含有read属性的为文件类型
                file_data[request_param] = param_value
            else:
                non_file_data[request_param] = param_value
        return non_file_data, file_data



# -*- coding: utf-8 -*-
"""
TencentBlueKing is pleased to support the open source community by making
蓝鲸智云 - Resource SDK (BlueKing - Resource SDK) available.
Copyright (C) 2023 THL A29 Limited,
a Tencent company. All rights reserved.
Licensed under the MIT License (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at https://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
either express or implied. See the License for the
specific language governing permissions and limitations under the License.
We undertake not to change the open source license (MIT license) applicable
to the current version of the project delivered to anyone in the future.
"""

import io
import os


def path_to_dotted(path):
    return ".".join([p for p in path.split(os.sep) if p])


def camel_to_underscore(camel_str):
    assert isinstance(camel_str, str)

    buf = io.StringIO()
    str_len = len(camel_str)

    for i in range(str_len):
        cur_letter = camel_str[i]
        if i and cur_letter == cur_letter.upper():
            prev_letter = camel_str[i - 1]
            next_letter = camel_str[i + 1] if i < str_len - 1 else "A"
            if cur_letter.isalpha():
                if prev_letter != prev_letter.upper() or next_letter != next_letter.upper():
                    buf.write("_")
        buf.write(cur_letter)

    result = buf.getvalue()
    buf.close()

    return result.lower()


def underscore_to_camel(underscore_str):
    assert isinstance(underscore_str, str)

    return "".join([x.capitalize() for x in underscore_str.split("_")])


invalid_filename_char = ["\\", "/", ":", "*", "?", '"', "<", ">", "|"]


def convert_filename(to_filename_string):
    r"""
    将文件名转换成符合linux文件系统规则的字符串
    需要转换的字符串：\ / : * ? " < > |
    :param to_filename_string:
    :return:
    """
    for index, c in enumerate(invalid_filename_char):
        target_string = "-%s-" % (chr(ord("a") + index))
        to_filename_string = to_filename_string.replace(c, target_string)

    return to_filename_string


output_string = list(map(convert_filename, invalid_filename_char))


def reconvert_filename(to_string_filename):
    """
    将转换过的文件名，重新变成之前的字符串
    :param to_string_filename:
    :return:
    """
    for index, c in enumerate(output_string):
        to_string_filename = to_string_filename.replace(c, invalid_filename_char[index])

    return to_string_filename


def cut_str_by_max_bytes(s: str, max_bytes: int, encoding: str = None):
    """
    字符串按最大字节长度裁剪
    """
    # 没有超过最大长度
    if len(s.encode(encoding=encoding) if encoding else s) <= max_bytes:
        return s

    # 逐字符计算字符串长度，查找截断字符串的位置
    new_s_len = 0
    max_index = -1
    for index, c in enumerate(s):
        c_bytes = c.encode(encoding=encoding) if encoding else c
        if len(c_bytes) + new_s_len > max_bytes:
            break
        new_s_len += len(c_bytes)
        max_index = index

    return s[: max_index + 1]



# -*- coding: utf-8 -*-
"""
TencentBlueKing is pleased to support the open source community by making
蓝鲸智云 - Resource SDK (BlueKing - Resource SDK) available.
Copyright (C) 2023 THL A29 Limited,
a Tencent company. All rights reserved.
Licensed under the MIT License (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at https://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
either express or implied. See the License for the
specific language governing permissions and limitations under the License.
We undertake not to change the open source license (MIT license) applicable
to the current version of the project delivered to anyone in the future.
"""

import json
import math
import os
from collections import OrderedDict

from django.utils.encoding import force_str
from django.utils.translation import gettext
from rest_framework import serializers
from rest_framework.fields import empty

from bk_resource.settings import bk_resource_settings
from bk_resource.utils.common_utils import ignored
from bk_resource.utils.logger import logger
from bk_resource.utils.text import camel_to_underscore


class FieldType(object):
    BOOLEAN = "Boolean"
    NUMBER = "Number"
    STRING = "String"
    INTEGER = "Integer"
    OBJECT = "Object"
    ARRAY = "Array"
    ENUM = "Enum"


def get_serializer_fields(serializer_class):
    """
    遍历serializer所有field，生成关于该serializer的schema列表
    """

    if not serializer_class:
        return []

    serializer = serializer_class()

    if isinstance(serializer, serializers.ListSerializer):
        return []

    if not isinstance(serializer, serializers.Serializer):
        return []

    fields = []
    for field in list(serializer.fields.values()):
        if field.read_only or isinstance(field, serializers.HiddenField):
            continue

        fields.append(field_to_schema(field))

    return fields


def field_to_schema(field):
    """
    根据serializer field生成关于该field数据结构的schema
    """
    description = force_str(field.label) if field.label else ""

    type_params = {
        "type": FieldType.STRING,
    }

    if isinstance(field, (serializers.ListSerializer, serializers.ListField)):
        child_schema = field_to_schema(field.child)
        type_params = {"type": FieldType.ARRAY, "items": child_schema}
    elif isinstance(field, serializers.Serializer):
        type_params = {
            "type": FieldType.OBJECT,
            "properties": OrderedDict([(key, field_to_schema(value)) for key, value in list(field.fields.items())]),
        }
    # elif isinstance(field, serializers.ManyRelatedField):
    #     type_params = {
    #         "type": FieldType.ARRAY,
    #         "items": FieldType.STRING,
    #     }
    elif isinstance(field, serializers.RelatedField):
        type_params = {
            "type": FieldType.STRING,
        }
    elif isinstance(field, (serializers.MultipleChoiceField, serializers.ChoiceField)):
        type_params = {
            "type": FieldType.ENUM,
            "choices": list(field.choices.keys()),
        }
    elif isinstance(field, serializers.BooleanField):
        type_params = {
            "type": FieldType.BOOLEAN,
        }
    elif isinstance(field, (serializers.DecimalField, serializers.FloatField)):
        type_params = {
            "type": FieldType.NUMBER,
        }
    elif isinstance(field, serializers.IntegerField):
        type_params = {
            "type": FieldType.INTEGER,
        }

    type_params.update(
        {
            "required": field.required,
            "name": field.field_name,
            "source_name": field.source,
            "description": description,
            "default": field.default,
        }
    )
    return type_params


def render_schema(fields, parent="", using_source=False):
    """
    将field schema渲染成apidoc的格式
    :param fields: schema list
    :param parent: field name of parent
    :param using_source: using source name of the field
    :return: list
    """
    print_list = []
    for field in fields:
        real_type = field["type"]
        field_name = field["source_name"] if using_source else field["name"]
        origin_name = "{}.{}".format(parent, field_name) if parent else field_name
        real_name = origin_name
        if field["type"] == FieldType.ARRAY:
            real_type = "%s[]" % field["items"]["type"]
        elif field["type"] == FieldType.ENUM:
            choices = ",".join(['"%s"' % c for c in field["choices"]])
            real_type = "{}={}".format(FieldType.STRING, choices)
        if field["default"] is not empty:
            real_name = "{}={}".format(real_name, field["default"])
        if not field["required"]:
            real_name = "[%s]" % real_name
        print_list.append("{{{}}} {} {}".format(real_type, real_name, field["description"]))

        if field["type"] == FieldType.ARRAY and field["items"]["type"] == FieldType.OBJECT:
            print_list += render_schema(
                fields=list(field["items"]["properties"].values()),
                parent=origin_name,
                using_source=using_source,
            )
        elif field["type"] == FieldType.OBJECT:
            print_list += render_schema(
                fields=list(field["properties"].values()),
                parent=origin_name,
                using_source=using_source,
            )
    return print_list


def get_underscore_viewset_name(viewset):
    return camel_to_underscore(viewset.__name__.replace("ViewSet", ""))


def _format_serializer_errors_core(errors, fields, params):
    """序列化器错误信息格式化"""
    for key, field_errors in list(errors.items()):
        label, sub_message = key, ""

        if key not in fields:
            sub_message = json.dumps(field_errors, ensure_ascii=False)
        else:
            field = fields[key]
            label = field.label or field.field_name
            if isinstance(field_errors, dict):
                if hasattr(field, "child"):
                    sub_format = _format_serializer_errors_core(field_errors, field.child.fields, params)
                else:
                    sub_format = _format_serializer_errors_core(field_errors, field.fields, params)
                sub_message += sub_format
            elif isinstance(field_errors, list):
                for error in field_errors:
                    # 若错误信息中有%s可将错误值加入其中
                    sub_message = error.format(**{key: params.get(key, "")})

        message = "({}) {}".format(label, sub_message)
        return message
    return ""


def format_serializer_errors(serializer):
    try:
        message = _format_serializer_errors_core(serializer.errors, serializer.fields, serializer.get_initial())
    except Exception as e:
        logger.warning(gettext("序列化器错误信息格式化失败，原因: %s"), e)
        return serializer.errors
    else:
        return message


def get_processes() -> int:
    """
    获取CPU数量或者容器内限制数量
    """

    # 环境变量
    processes = bk_resource_settings.RESOURCE_BULK_REQUEST_PROCESSES
    if processes and isinstance(processes, int):
        return processes

    # CPU
    cpu_count = os.cpu_count()

    # 容器限制
    # 先给默认值避免后续比较报错
    container_cpu = cpu_count
    with ignored(Exception, log_exception=False):
        with open("/sys/fs/cgroup/cpu/cpu.cfs_quota_us", "r") as f:
            cfs_quota_us = float(f.read().strip())
        with open("/sys/fs/cgroup/cpu/cpu.cfs_period_us", "r") as f:
            cfs_period_us = float(f.read().strip())
        # 当 cfs_quota_us = -1 时表示无限制
        if cfs_quota_us > 0:
            # 向上取整，至少保证有 1 个
            container_cpu = math.ceil(cfs_quota_us / cfs_period_us)

    # 取限制中的较小值
    return min(container_cpu, cpu_count)



# -*- coding: utf-8 -*-
"""
TencentBlueKing is pleased to support the open source community by making
蓝鲸智云 - Resource SDK (BlueKing - Resource SDK) available.
Copyright (C) 2023 THL A29 Limited,
a Tencent company. All rights reserved.
Licensed under the MIT License (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at https://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
either express or implied. See the License for the
specific language governing permissions and limitations under the License.
We undertake not to change the open source license (MIT license) applicable
to the current version of the project delivered to anyone in the future.
"""

import abc
import json
from typing import Union

import arrow
from django.db import models
from django.http.response import HttpResponseBase
from django.utils.translation import gettext
from rest_framework.response import Response

from bk_resource.exceptions import ValidateException
from bk_resource.settings import bk_resource_settings
from bk_resource.tasks import run_perform_request
from bk_resource.tools import format_serializer_errors, get_processes
from bk_resource.utils.logger import logger
from bk_resource.utils.request import get_request_username
from bk_resource.utils.thread_backend import ThreadPool

__doc__ = """
Non-ORM for DRF 的架构：
serializers -> resources -> views

各模块的职责
serializers: 负责请求数据与返回数据的校验，包括简单的数据渲染
resources: 此处编写业务逻辑（工作重点）
views: 定义请求方法、使用到的Resource、是否分页、鉴权等配置（几乎没有代码量）

Resource定义：
对传入的数据进行特定处理并返回处理后的数据，这样的处理单元称为Resource
可以通过继承Resource类来实现自定义的Resource

Resource的执行流程：
1. 新建Resource实例
2. 若用户未手动设置需要用到的RequestSerializer及ResponseSerializer，
   则根据命名规则自动查找可用的serializers，并进行设置
3. 调用request方法，并传入请求参数(request_data)
4. 调用request_serializer(如果有)对request_data进行数据校验，
   返回validated_request_data
5. 用户自行实现perform_request方法，根据validated_request_data，通过用户编写的业务
   逻辑，生成返回结果response_data
6. 调用response_serializer(如果有)对response_data进行数据校验，
   返回validated_response_data
7. 将validated_response_data交给上层的views，由views进行最后的数据渲染及返回到前端

为了Resource能够正确搜索到其对应的serializers，需要遵循以下命名规则：
某Resource名称为: TestResource
其对应的request serializer应命名为: TestRequestSerializer
其对应的response serializer应命名为: TestResponseSerializer

"""


class Empty:
    ...


class Resource(metaclass=abc.ABCMeta):
    RequestSerializer = None
    ResponseSerializer = None
    serializer_class = None
    _request_serializer = Empty()
    _response_serializer = Empty()

    # 提供一个serializers模块，在实例化某个Resource时，在该模块内自动查找符合命名
    # 规则的serializers，并进行自动配置
    serializers_module = None

    # 数据是否为对象的列表
    many_request_data = False
    many_response_data = False

    # 支持记录请求参数(settings开启：ENABLE_RESOURCE_DATA_COLLECT后，
    # 记录所有`support_data_collect`为True的resource请求)
    support_data_collect = True

    # 绑定 request 对象到请求参数中
    bind_request = False

    # swagger扩展信息
    name = ""
    tags = []

    def __init__(self, context=None):
        (
            self.RequestSerializer,
            self.ResponseSerializer,
        ) = self._search_serializer_class()

        self.context = context
        self._task_manager = None

    def __call__(self, *args, **kwargs):
        # thread safe
        resource = self.__class__()
        # 如果没有配置则退出
        if not resource.support_data_collect:
            data = resource.request(*args, **kwargs)
            assert not isinstance(data, Response) and not isinstance(data, HttpResponseBase), (
                gettext("[%s] 响应体为Response，只允许在ViewSet调用") % resource.__class__.__name__
            )
            return data

        resource_name = "{}.{}".format(resource.__class__.__module__, resource.__class__.__name__)
        start_time = arrow.now().datetime
        request_data = {"args": args, "kwargs": kwargs}
        response_data = ""
        try:
            response_data = resource.request(*args, **kwargs)
            assert not isinstance(response_data, Response), (
                gettext("[%s] 响应体为Response，只允许在ViewSet调用") % resource.__class__.__name__
            )
            return response_data
        except Exception as err:
            logger.exception(err)
            response_data = str(err)
            raise err
        finally:
            end_time = arrow.now().datetime
            # 记录请求日志
            log_handler = bk_resource_settings.REQUEST_LOG_HANDLER
            log_handler(resource_name, start_time, end_time, request_data, response_data).record()

    def _get_data_string(self, data):
        """
        数据转换为字符串，同时避免QuerySet查询
        """
        if isinstance(data, (dict, list, str, int)) or data is None:
            try:
                data_string = json.dumps(data, ensure_ascii=False)
            except Exception as e:
                data_string = str(e)
        else:
            data_string = str(type(data))
        return data_string

    @property
    def request_serializer(self):
        """
        :rtype: serializers.Serializer
        """
        if isinstance(self._request_serializer, Empty):
            msg = "You must call `.validate_request_data()` before accessing `.request_serializer`."
            raise AssertionError(msg)
        return self._request_serializer

    @property
    def response_serializer(self):
        """
        :rtype: serializers.Serializer
        """
        if isinstance(self._response_serializer, Empty):
            msg = "You must call `.validate_response_data()` before accessing `.response_serializer`."
            raise AssertionError(msg)
        return self._response_serializer

    @classmethod
    def get_resource_name(cls):
        subclass_name = cls.__name__
        resource_name = subclass_name.replace("Resource", "")
        return resource_name

    @classmethod
    def _search_serializer_class(cls):
        """
        搜索该Resource对应的两个Serializer
        """

        # 若类的内部声明了RequestSerializer和ResponseSerializer，则优先使用
        request_serializer_class = cls.RequestSerializer
        response_serializer_class = cls.ResponseSerializer

        # 若类中没有声明，则对指定模块进行搜索
        resource_name = cls.get_resource_name()
        request_serializer_class_name = "%sRequestSerializer" % resource_name
        response_serializer_class_name = "%sResponseSerializer" % resource_name

        if cls.serializers_module:
            if not request_serializer_class:
                request_serializer_class = getattr(cls.serializers_module, request_serializer_class_name, None)

            if not response_serializer_class:
                response_serializer_class = getattr(cls.serializers_module, response_serializer_class_name, None)

        return request_serializer_class, response_serializer_class

    @abc.abstractmethod
    def perform_request(self, validated_request_data):
        """
        此处为Resource的业务逻辑，由子类实现
        将request_data通过一定的逻辑转化为response_data

        example:
            return validated_request_data
        """
        raise NotImplementedError

    def validate_request_data(self, request_data):
        """
        校验请求数据
        """
        self._request_serializer = None
        if not self.RequestSerializer:
            return request_data

        # model类型的数据需要特殊处理
        if isinstance(request_data, (models.Model, models.QuerySet)):
            request_serializer = self.RequestSerializer(request_data, many=self.many_request_data)
            self._request_serializer = request_serializer
            return request_serializer.data
        else:
            request_serializer = self.RequestSerializer(data=request_data, many=self.many_request_data)
            self._request_serializer = request_serializer
            is_valid_request = request_serializer.is_valid()
            if not is_valid_request:
                msg = gettext("Resource[%s] 请求参数格式错误：%s") % (
                    self.get_resource_name(),
                    format_serializer_errors(request_serializer),
                )
                logger.error(msg)
                raise ValidateException(msg)
            return request_serializer.validated_data

    def validate_response_data(self, response_data):
        """
        校验返回数据
        """
        self._response_serializer = None
        if not self.ResponseSerializer:
            return response_data

        # model类型的数据需要特殊处理
        if isinstance(response_data, (models.Model, models.QuerySet)):
            response_serializer = self.ResponseSerializer(response_data, many=self.many_response_data)
            self._response_serializer = response_serializer
            return response_serializer.data
        else:
            response_serializer = self.ResponseSerializer(data=response_data, many=self.many_response_data)
            self._response_serializer = response_serializer
            is_valid_response = response_serializer.is_valid()
            if not is_valid_response:
                msg = gettext("Resource[%s] 返回参数格式错误：%s") % (
                    self.get_resource_name(),
                    format_serializer_errors(response_serializer),
                )
                logger.error(msg)
                raise ValidateException(msg)
            return response_serializer.validated_data

    def build_extra_params(self, request_data: Union[models.Model, dict], validated_request_data: dict) -> dict:
        """
        ModelResource补全参数埋点
        """
        return validated_request_data

    def request(self, request_data=None, **kwargs):
        """
        执行请求，并对请求数据和返回数据进行数据校验
        """
        request_data = request_data or kwargs
        validated_request_data = self.validate_request_data(request_data)
        validated_request_data = self.build_extra_params(request_data, validated_request_data)

        # 注入request
        if kwargs.get("_request") and not validated_request_data.get("_request"):
            validated_request_data["_request"] = kwargs["_request"]

        response_data = self.perform_request(validated_request_data)

        validated_response_data = self.validate_response_data(response_data)

        return validated_response_data

    def bulk_request(self, request_data_iterable=None, ignore_exceptions=False):
        """
        基于多线程的批量并发请求
        """

        # 预检查
        if not isinstance(request_data_iterable, (list, tuple)):
            raise TypeError("'request_data_iterable' object is not iterable")

        # 模块引入，放在文件头可能导致 django 未完全初始化异常
        from blueapps.utils.request_provider import get_local_request

        # 初始化
        futures = []
        _request = get_local_request()

        # 线程池
        with ThreadPool(processes=get_processes()) as pool:
            for request_data in request_data_iterable:
                futures.append(pool.apply_async(self.request, args=(request_data,), kwds={"_request": _request}))
            pool.close()
            pool.join()

        # 获取结果
        results = []
        exceptions = []
        for future in futures:
            try:
                results.append(future.get())
            except Exception as e:
                # 判断是否忽略错误
                if not ignore_exceptions:
                    raise e
                exceptions.append(e)
                results.append(None)

        # 如果全部报错，则必须抛出错误
        if exceptions and len(exceptions) == len(futures):
            raise exceptions[0]

        return results

    def update_state(self, state, message=None, data=None):
        """
        更新执行状态
        """
        state_message = "Async resource task running - {} [state=`{}` message=`{}` data=`{}`]".format(
            self.get_resource_name(), state, message, data
        )
        logger.info(state_message)

        if not self._task_manager:
            return

        meta = {
            "message": message,
            "data": data,
        }
        self._task_manager.update_state(state=state, meta=meta)

    def delay(self, request_data=None, **kwargs):
        """
        执行celery异步任务
        """
        request_data = request_data or kwargs
        return self.apply_async(request_data)

    def apply_async(self, request_data, **kwargs):
        """
        执行celery异步任务（高级）
        """
        async_task = run_perform_request.apply_async(
            args=(f"{self.__module__}.{self.__class__.__name__}", get_request_username(), request_data), **kwargs
        )
        return {"task_id": async_task.id}

    @classmethod
    def etag(cls, request, *args, **kwargs):
        return

    @classmethod
    def last_modified(cls, request, *args, **kwargs):
        return



# -*- coding: utf-8 -*-
"""
TencentBlueKing is pleased to support the open source community by making
蓝鲸智云 - Resource SDK (BlueKing - Resource SDK) available.
Copyright (C) 2023 THL A29 Limited,
a Tencent company. All rights reserved.
Licensed under the MIT License (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at https://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
either express or implied. See the License for the
specific language governing permissions and limitations under the License.
We undertake not to change the open source license (MIT license) applicable
to the current version of the project delivered to anyone in the future.
"""

from collections import defaultdict
from typing import Union

from django.conf import settings
from django.utils.module_loading import import_string


class BkResourceSettings:
    """
    Settings
    """

    SETTINGS = defaultdict(
        None,
        DEFAULT_API_DIR="api",
        DEFAULT_RESOURCE_DIRS=[],
        LOCAL_CACHE_ENABLE=False,
        INTERFACE_COMMON_PARAMS={
            "bk_app_code": settings.APP_CODE,
            "bk_app_secret": settings.SECRET_KEY,
        },
        DEFAULT_ERROR_RESPONSE_SERIALIZER="bk_resource.serializers.ErrorResponseSerializer",
        DEFAULT_PAGINATOR_RESPONSE_BUILDER="bk_resource.serializers.PaginatorResponseBuilder",
        DEFAULT_STANDARD_RESPONSE_BUILDER="bk_resource.serializers.StandardResponseBuilder",
        DEFAULT_SWAGGER_SCHEMA_CLASS="bk_resource.utils.inspectors.BkResourceSwaggerAutoSchema",
        REQUEST_LOG_HANDLER="bk_resource.utils.request_log.RequestLogHandler",
        REQUEST_LOG_SPLIT_LENGTH=0,
        REQUEST_VERIFY=True,
        PLATFORM_AUTH_ENABLED=False,
        PLATFORM_AUTH_ACCESS_TOKEN=None,
        PLATFORM_AUTH_ACCESS_USERNAME=None,
        REQUEST_BKAPI_COOKIE_FIELDS=["blueking_language", "django_language"],
        RESOURCE_BULK_REQUEST_PROCESSES=None,
    )

    LAZY_IMPORT_SETTINGS = (
        "DEFAULT_ERROR_RESPONSE_SERIALIZER",
        "DEFAULT_PAGINATOR_RESPONSE_BUILDER",
        "DEFAULT_STANDARD_RESPONSE_BUILDER",
        "DEFAULT_SWAGGER_SCHEMA_CLASS",
        "REQUEST_LOG_HANDLER",
    )

    LOADED_SETTINGS = {}

    def __getattr__(self, key: str) -> any:
        """
        get settings from self
        """

        # 0. validate key
        self.__validate_key(key)

        # 1. load Key from cache
        if key in self.LOADED_SETTINGS.keys():
            return self.LOADED_SETTINGS[key]

        # 2. load key from django.conf.settings
        if key in self.custom_settings.keys():
            return self.__load_key(key, self.custom_settings[key])

        # 3. load key from default settings
        return self.__load_key(key, self.SETTINGS[key])

    def __validate_key(self, key: str) -> None:
        """
        validate key
        """

        if key not in self.SETTINGS.keys():
            raise KeyError("[%s] is not a valid key for bk_resource" % key)

    def __load_key(self, key: str, val: Union[str, callable]) -> any:
        """
        load lazy import key to cache
        """

        # load lazy setting
        if key in self.LAZY_IMPORT_SETTINGS and isinstance(val, str):
            val = import_string(val)
            self.LOADED_SETTINGS[key] = val

        # return value
        return val

    @property
    def custom_settings(self) -> dict:
        """
        Custom settings in django.conf.settings
        """

        return getattr(settings, "BK_RESOURCE", {})


bk_resource_settings = BkResourceSettings()



# -*- coding: utf-8 -*-
"""
TencentBlueKing is pleased to support the open source community by making
蓝鲸智云 - Resource SDK (BlueKing - Resource SDK) available.
Copyright (C) 2023 THL A29 Limited,
a Tencent company. All rights reserved.
Licensed under the MIT License (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at https://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
either express or implied. See the License for the
specific language governing permissions and limitations under the License.
We undertake not to change the open source license (MIT license) applicable
to the current version of the project delivered to anyone in the future.
"""

import abc
import json
from datetime import datetime

from bk_resource.settings import bk_resource_settings
from bk_resource.utils.logger import logger


class BaseRequestLogHandler:
    def __init__(
        self,
        resource_name: str,
        start_time: datetime,
        end_time: datetime,
        request_data: any,
        response_data: any,
    ):
        self.resource_name = resource_name
        self.start_time = start_time
        self.end_time = end_time
        self.request_data = self.parse_json(request_data)
        self.response_data = self.parse_json(response_data)
        if bk_resource_settings.REQUEST_LOG_SPLIT_LENGTH:
            log_length = len(self.response_data)
            self.response_data = self.response_data[: bk_resource_settings.REQUEST_LOG_SPLIT_LENGTH] + (
                "." * 6 if log_length > bk_resource_settings.REQUEST_LOG_SPLIT_LENGTH else ""
            )

    @abc.abstractmethod
    def record(self):
        ...

    @classmethod
    def parse_json(cls, data: dict) -> str:
        try:
            return json.dumps(data, ensure_ascii=False)
        except Exception:
            return str(data)


class RequestLogHandler(BaseRequestLogHandler):
    def record(self):
        msg = (
            "[ResourceRequestLog]\n"
            "AppCode => %s\n"
            "Username => %s\n"
            "Resource => %s\n"
            "StartTime => %s\n"
            "EndTime => %s\n"
            "RequestData => %s\n"
            "ResponseData => %s"
        )
        logger.info(
            msg,
            self.get_app_code(),
            self.get_username(),
            self.resource_name,
            self.start_time,
            self.end_time,
            self.request_data,
            self.response_data,
        )

    @classmethod
    def get_username(cls) -> str:
        """获取请求用户名"""
        from blueapps.utils.request_provider import get_local_request

        try:
            return get_local_request().user.username
        except (IndexError, AttributeError):
            return ""

    @classmethod
    def get_app_code(cls) -> str:
        """获取AppCode"""
        from blueapps.utils.request_provider import get_local_request

        try:
            app = get_local_request().app
            return f"{app.bk_app_code}{'' if app.verified else '(unverified)'}"
        except (IndexError, AttributeError):
            return ""



# -*- coding: utf-8 -*-
"""
TencentBlueKing is pleased to support the open source community by making
蓝鲸智云 - Resource SDK (BlueKing - Resource SDK) available.
Copyright (C) 2023 THL A29 Limited,
a Tencent company. All rights reserved.
Licensed under the MIT License (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at https://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
either express or implied. See the License for the
specific language governing permissions and limitations under the License.
We undertake not to change the open source license (MIT license) applicable
to the current version of the project delivered to anyone in the future.
"""

from _thread import get_ident
from contextlib import contextmanager

__all__ = ["local", "Local", "get_ident"]


class Localbase(object):
    __slots__ = ("__storage__", "__ident_func__")

    def __new__(cls, *args, **kwargs):
        self = object.__new__(cls)
        object.__setattr__(self, "__storage__", {})
        object.__setattr__(self, "__ident_func__", get_ident)
        return self


class Local(Localbase):
    def __iter__(self):
        ident = self.__ident_func__()
        try:
            return iter(list(self.__storage__[ident].items()))
        except KeyError:
            return iter([])

    def __release_local__(self):
        self.__storage__.pop(self.__ident_func__(), None)

    def __getattr__(self, name):
        ident = self.__ident_func__()
        try:
            return self.__storage__[ident][name]
        except KeyError:
            raise AttributeError(name)

    def __setattr__(self, name, value):
        if name in ("__storage__", "__ident_func__"):
            raise AttributeError("{!r} object attribute '{}' is read-only".format(self.__class__.__name__, name))

        ident = self.__ident_func__()
        storage = self.__storage__
        if ident not in storage:
            storage[ident] = dict()
        storage[ident][name] = value

    def __delattr__(self, name):
        if name in ("__storage__", "__ident_func__"):
            raise AttributeError("{!r} object attribute '{}' is read-only".format(self.__class__.__name__, name))

        ident = self.__ident_func__()
        try:
            del self.__storage__[ident][name]
            if len(self.__storage__[ident]) == 0:
                self.__release_local__()
        except KeyError:
            raise AttributeError(name)

    def clear(self):
        self.__release_local__()


local = Local()


@contextmanager
def with_request_local():
    local_vars = {}
    for k in ["username", "current_request"]:
        if hasattr(local, k):
            local_vars[k] = getattr(local, k)
            delattr(local, k)

    try:
        yield local
    finally:
        for k, v in list(local_vars.items()):
            setattr(local, k, v)


@contextmanager
def with_client_user(username):
    with with_request_local() as local:
        local.username = username
        yield


@contextmanager
def with_client_operator(update_user):
    with with_request_local() as local:
        local.operator = update_user
        yield



# -*- coding: utf-8 -*-
"""
TencentBlueKing is pleased to support the open source community by making
蓝鲸智云 - Resource SDK (BlueKing - Resource SDK) available.
Copyright (C) 2023 THL A29 Limited,
a Tencent company. All rights reserved.
Licensed under the MIT License (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at https://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
either express or implied. See the License for the
specific language governing permissions and limitations under the License.
We undertake not to change the open source license (MIT license) applicable
to the current version of the project delivered to anyone in the future.
"""

from typing import List

import arrow
from django.core.exceptions import ObjectDoesNotExist
from django.http.response import HttpResponseBase
from django.utils.decorators import method_decorator
from django.utils.translation import gettext
from django.views.decorators.cache import cache_control
from drf_yasg.utils import swagger_auto_schema
from rest_framework import viewsets
from rest_framework.decorators import action
from rest_framework.response import Response
from rest_framework.serializers import Serializer
from rest_framework_condition import condition

from bk_resource.base import Resource
from bk_resource.settings import bk_resource_settings


class ResourceRoute(object):
    """
    Resource的视图配置，应用于viewsets
    """

    def __init__(
        self,
        method,
        resource_class,
        endpoint="",
        pk_field=None,
        enable_paginate=False,
        content_encoding=None,
        decorators=None,
    ):
        """
        :param method: 请求方法，目前仅支持GET和POST
        :param resource_class: 所用到的Resource类
        :param endpoint: 端点名称，不提供则为list或create
        :param pk_field: 主键名称，如果不为空，则该视图为 detail route
        :param enable_paginate: 是否对结果进行分页
        :param content_encoding: 返回数据内容编码类型
        :params decorators: 给view_func添加的装饰器列表
        """
        # if method.upper() not in ["GET", "POST"]:
        #     raise ValueError(_("method参数错误，目前仅支持GET或POST方法"))

        self.method = method.upper()

        if isinstance(resource_class, Resource):
            resource_class = resource_class.__class__
        if not issubclass(resource_class, Resource):
            raise ValueError(gettext("resource_class参数必须提供Resource的子类, 当前类型: %s") % resource_class)

        self.resource_class = resource_class

        self.endpoint = endpoint

        self.enable_paginate = enable_paginate

        self.content_encoding = content_encoding

        self.pk_field = pk_field

        self.decorators = decorators if isinstance(decorators, list) else None


class ResourceViewSet(viewsets.GenericViewSet):
    EMPTY_ENDPOINT_METHODS = {
        "GET": "list",
        "POST": "create",
        "PUT": "update",
        "PATCH": "partial_update",
        "DELETE": "destroy",
    }

    # 用于执行请求的Resource类
    resource_routes: List[ResourceRoute] = []
    filter_backends = []

    # swagger
    if bk_resource_settings.DEFAULT_SWAGGER_SCHEMA_CLASS is not None:
        swagger_schema = bk_resource_settings.DEFAULT_SWAGGER_SCHEMA_CLASS

    def get_serializer_class(self):
        """
        获取序列化器
        """
        serializer_class = None
        for route in self.resource_routes:
            if self.action == route.endpoint:
                serializer_class = route.resource_class.RequestSerializer
                break

        if not serializer_class:
            serializer_class = Serializer

        class Meta:
            ref_name = None

        serializer_class.Meta = Meta
        return serializer_class

    def get_queryset(self):
        """
        添加默认函数，避免swagger生成报错
        """
        return

    @classmethod
    def generate_endpoint(cls):
        for resource_route in cls.resource_routes:
            # 生成方法模版
            function = cls._generate_function_template(resource_route)

            # 请求序列化
            request_serializer_class = resource_route.resource_class.RequestSerializer or Serializer
            request_serializer = request_serializer_class(many=resource_route.resource_class.many_request_data)

            # 响应序列化
            response_serializer_class = (
                resource_route.resource_class.ResponseSerializer
                or resource_route.resource_class.serializer_class
                or Serializer
            )
            # 支持使用 Field 作为 data 内容
            try:
                response_serializer = response_serializer_class(many=resource_route.resource_class.many_response_data)
            except TypeError:
                response_serializer = response_serializer_class()

            # 启用分页
            if resource_route.enable_paginate:
                paginator_response_builder = bk_resource_settings.DEFAULT_PAGINATOR_RESPONSE_BUILDER
                response_serializer = paginator_response_builder(
                    resource_class=resource_route.resource_class,
                    data_serializer=response_serializer,
                ).serializer

            # 统一响应格式
            standard_response_builder = bk_resource_settings.DEFAULT_STANDARD_RESPONSE_BUILDER
            response_serializer = standard_response_builder(
                resource_class=resource_route.resource_class,
                data_serializer=response_serializer,
            ).serializer

            decorator_function = swagger_auto_schema(
                responses={
                    200: response_serializer,
                    500: bk_resource_settings.DEFAULT_ERROR_RESPONSE_SERIALIZER,
                },
                operation_description=resource_route.resource_class.__doc__,
                request_body=request_serializer if resource_route.method in ["POST", "PUT", "DELETE"] else None,
                query_serializer=request_serializer if resource_route.method == "GET" else None,
                operation_summary=getattr(resource_route.resource_class, "name", None),
                tags=getattr(resource_route.resource_class, "tags", None),
                enable_paginator=resource_route.enable_paginate,
            )

            # 添加装饰器
            if resource_route.decorators:
                for decorator in resource_route.decorators:
                    function = decorator(function)

            # 为Viewset设置方法
            if not resource_route.endpoint:
                function = decorator_function(function)
                # 默认方法无需加装饰器，否则会报错
                if resource_route.method == "GET":
                    if resource_route.pk_field:
                        cls.retrieve = function
                    else:
                        cls.list = function
                elif resource_route.method == "POST":
                    if resource_route.pk_field:
                        raise AssertionError(
                            gettext("当请求方法为 %s，且 endpoint 为空时，禁止设置 pk_field 参数") % resource_route.method
                        )
                    cls.create = function
                elif resource_route.method in cls.EMPTY_ENDPOINT_METHODS:
                    if not resource_route.pk_field:
                        raise AssertionError(
                            gettext("当请求方法为 %s，且 endpoint 为空时，必须提供 pk_field 参数") % resource_route.method
                        )
                    setattr(cls, cls.EMPTY_ENDPOINT_METHODS[resource_route.method], function)
                else:
                    raise AssertionError(gettext("不支持的请求方法: %s，请确认resource_routes配置是否正确!") % resource_route.method)
            else:
                function = method_decorator(cache_control(max_age=0, private=True))(function)
                function.__name__ = resource_route.endpoint
                if resource_route.pk_field:
                    function = action(methods=[resource_route.method], detail=True)(function)
                else:
                    function = action(methods=[resource_route.method], detail=False)(function)
                function = condition(
                    etag_func=resource_route.resource_class.etag,
                    last_modified_func=resource_route.resource_class.last_modified,
                )(function)

                function = decorator_function(function)
                setattr(cls, resource_route.endpoint, function)

    @classmethod
    def _generate_function_template(cls, resource_route: ResourceRoute):
        """
        生成方法模版
        """

        def template(self, request, *args, **kwargs):
            start_time = arrow.now().datetime

            resource = resource_route.resource_class()
            request_data = request.query_params.copy() if resource_route.method == "GET" else request.data

            # 如果是detail route，需要重url参数中获取主键，并塞到请求参数中
            if resource_route.pk_field:
                request_data.update({resource_route.pk_field: kwargs[cls.lookup_field]})

            # 合并url path params
            if kwargs:
                request_data.update(kwargs)

            params = {"request_data": request_data}

            # 判断是否需要绑定request对象
            if resource.bind_request:
                params["_request"] = request

            is_async_task = "HTTP_X_ASYNC_TASK" in request.META
            if is_async_task:
                # 执行异步任务
                data = resource.delay(**params)
                response = Response(data)
            else:
                try:
                    data = resource.request(**params)
                    if isinstance(data, Response):
                        response = data
                        data = data.data
                    elif isinstance(data, HttpResponseBase):
                        response = data
                        data = getattr(data, "content", "")
                    elif resource_route.enable_paginate:
                        page = self.paginate_queryset(data)
                        response = self.get_paginated_response(page)
                    else:
                        response = Response(data)
                except ObjectDoesNotExist:
                    # 默认异常，有特殊情况在程序逻辑中处理
                    raise Exception(gettext("[%s] 访问的资源不存在") % resource_route.resource_class.__name__)

            if resource_route.content_encoding:
                response.content_encoding = resource_route.content_encoding

            end_time = arrow.now().datetime

            # 不记录日志的直接返回
            if not resource.support_data_collect:
                return response

            # 记录请求日志
            resource_name = "{}.{}".format(resource.__class__.__module__, resource.__class__.__name__)
            log_handler = bk_resource_settings.REQUEST_LOG_HANDLER
            log_handler(resource_name, start_time, end_time, request_data, data).record()

            return response

        return template



# -*- coding: utf-8 -*-
"""
TencentBlueKing is pleased to support the open source community by making
蓝鲸智云 - Resource SDK (BlueKing - Resource SDK) available.
Copyright (C) 2023 THL A29 Limited,
a Tencent company. All rights reserved.
Licensed under the MIT License (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at https://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
either express or implied. See the License for the
specific language governing permissions and limitations under the License.
We undertake not to change the open source license (MIT license) applicable
to the current version of the project delivered to anyone in the future.
"""

from rest_framework.routers import DefaultRouter
from rest_framework.viewsets import GenericViewSet

from bk_resource.tools import get_underscore_viewset_name
from bk_resource.viewsets import ResourceViewSet


class ResourceRouter(DefaultRouter):
    @staticmethod
    def _init_resource_viewset(viewset):
        """
        初始化ResourceViewset，动态增加方法，在register router前必须进行调用
        @:param viewset: ResourceViewset类
        """
        if isinstance(viewset, type) and issubclass(viewset, ResourceViewSet):
            viewset.generate_endpoint()

    def register(self, prefix, viewset, base_name=None):
        """
        注册单个ResourceViewset
        """
        self._init_resource_viewset(viewset)
        super(ResourceRouter, self).register(prefix, viewset, base_name)

    def register_module(self, viewset_module, prefix=None):
        """
        注册整个ResourceViewset模块，并根据类的命名规则自动生成对应的url
        """
        for attr, viewset in viewset_module.__dict__.items():
            # 全小写的属性不是类，忽略
            if attr.startswith("_") or attr[0].islower():
                continue

            if isinstance(viewset, type) and issubclass(viewset, GenericViewSet):
                module_prefix = self.get_default_basename(viewset)
                if prefix:
                    module_prefix = f"{prefix}/{module_prefix}" if module_prefix else prefix
                self.register(module_prefix, viewset)

    def get_default_basename(self, viewset):
        return get_underscore_viewset_name(viewset)



# -*- coding: utf-8 -*-
"""
TencentBlueKing is pleased to support the open source community by making
蓝鲸智云 - Resource SDK (BlueKing - Resource SDK) available.
Copyright (C) 2023 THL A29 Limited,
a Tencent company. All rights reserved.
Licensed under the MIT License (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at https://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
either express or implied. See the License for the
specific language governing permissions and limitations under the License.
We undertake not to change the open source license (MIT license) applicable
to the current version of the project delivered to anyone in the future.
"""

from django.test import RequestFactory

from bk_resource.utils.local import local


def get_request_username(default=""):
    try:
        from blueapps.utils.request_provider import get_local_request

        username = get_local_request().user.username
    except Exception:  # pylint: disable=broad-except
        username = get_local_username()
        if not username:
            username = default
    return username


def get_local_username():
    """从local对象中获取用户信息（celery）"""
    return getattr(local, "username", None)


def set_local_username(username):
    local.username = username


def get_moke_request(**kwargs):
    return RequestFactory().request(**kwargs)



# -*- coding: utf-8 -*-
"""
TencentBlueKing is pleased to support the open source community by making
蓝鲸智云 - Resource SDK (BlueKing - Resource SDK) available.
Copyright (C) 2023 THL A29 Limited,
a Tencent company. All rights reserved.
Licensed under the MIT License (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at https://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
either express or implied. See the License for the
specific language governing permissions and limitations under the License.
We undertake not to change the open source license (MIT license) applicable
to the current version of the project delivered to anyone in the future.
"""

import inspect
from contextlib import contextmanager
from importlib import import_module

from django.conf import settings
from django.utils.module_loading import import_string

from bk_resource.base import Resource
from bk_resource.management.exceptions import (
    ResourceModuleConflict,
    ResourceModuleNotRegistered,
    ResourceNotRegistered,
)
from bk_resource.management.finder import API_DIR, ResourceFinder
from bk_resource.utils.logger import logger
from bk_resource.utils.text import camel_to_underscore, path_to_dotted

__setup__ = False
__doc__ = """
自动发现项目下resource和adapter和api
    cc
    ├── adapter
    │   ├── default.py
    │   ├── community
    │   │   └── resources.py
    │   └── enterprise
    │       └── resources.py
    └── resources.py
    使用:
        # api: 代表基于ESB/APIGW调用的接口调用
        # api.$module.$api_name
            api.bkdata.query_data -> api/bkdata/default.py: QueryDataResource
        # resource: 基于业务逻辑的封装
        resource.plugin -> plugin/resources.py
            resource.plugin.install_plugin -> plugin/resources.py: InstallPluginResource
        # adapter: 针对不同版本的逻辑差异进行的封装
        adapter.cc -> cc/adapter/default.py -> cc/adapter/${platform}/resources.py
        # 调用adapter.cc 即可访问对应文件下的resource，
        # 如果在${platform}/resources.py里面有相同定义，会重载default.py下的resource
    """


@contextmanager
def _lazy_load(_instance):
    if not _instance.loaded:
        getattr(_instance, "_setup", lambda: None)()
    yield _instance


def lazy_load(func):
    def wrapper(resource_module_instance, *args, **kwargs):
        with _lazy_load(resource_module_instance):
            return func(resource_module_instance, *args, **kwargs)

    return wrapper


class ResourceShortcut(object):

    _package_pool = {}

    _entry = "resources"

    def __new__(cls, module_path):
        def init_instance(self, m_path):
            if not m_path.endswith(self._entry):
                m_path += ".{}".format(self._entry)
            self._path = m_path
            self._package = None
            self._methods = {}
            self.loaded = False

        if module_path not in cls._package_pool:
            instance = object.__new__(cls)
            init_instance(instance, module_path)
            cls._package_pool[module_path] = instance
        return cls._package_pool[module_path]

    def __init__(self, module_path):
        # init in __new__ function
        pass

    def __getitem__(self, x):
        # delete _entry(replace once from right)
        path = "".join(self._path.rsplit(self._entry, 1)).strip(".")
        dotted_path = path.split(".")
        if not isinstance(x, slice):
            return dotted_path[x]
        rm = resource
        while dotted_path[:-1]:
            rm = getattr(rm, dotted_path.pop(0))
        return tuple(rm)

    def _setup(self):
        try:
            self._package = import_module(self._path)
        except ImportError as e:
            raise ImportError("resource called before {} setup. detail: {}".format(self._path, e))
        for name, obj in list(self._package.__dict__.items()):
            if name.startswith("_"):
                continue

            if inspect.isabstract(obj):
                continue

            if inspect.isclass(obj) and issubclass(obj, Resource):
                cleaned_name = "".join(name.rsplit("Resource", 1)) if name.endswith("Resource") else name
                property_name = camel_to_underscore(cleaned_name)
                setattr(self, property_name, obj())
                self._methods[property_name] = obj

            if inspect.isfunction(obj):
                setattr(self, name, obj)
                self._methods[name] = obj

        self.loaded = True

    @lazy_load
    def __getattr__(self, item):
        if item in self._methods:
            return getattr(self, item)
        try:
            return ResourceShortcut(import_module("{}.{}".format(self._path, item)).__name__)
        except ImportError:
            try:
                return import_string("{}.{}".format(self._path, item))
            except ImportError:
                raise ResourceNotRegistered("Resource {} not in [{}]".format(item, self._package.__name__))

    @lazy_load
    def list_method(self):
        return list(self._methods.keys())

    def reload_method(self, method, func):
        setattr(self, method, func)
        self._methods[method] = func


class AdapterResourceShortcut(ResourceShortcut):

    _entry = "adapter.default"
    _package_pool = {}


class APIResourceShortcut(ResourceShortcut):
    _entry = "default"
    _package_pool = {}


class ResourceManager(tuple):
    __parent__ = None

    def __contains__(self, item):
        return item is not None and (self is item or item[: len(self)] == self)

    def __getattr__(self, name):
        if __setup__:
            got = getattr(self.transform(), name, None)
            if got is None:
                raise ResourceModuleNotRegistered(
                    'module: "%s" is not registered, maybe not in `INSTALLED_APPS` ?' % name
                )
            return got

        new = ResourceManager(self + (name,))
        setattr(self, name, new)
        new.__parent__ = self
        return new

    def __repr__(self):
        return "resource" + ("." if self else "") + ".".join(self)

    def __bool__(self):
        return bool(len(self))

    @property
    def __root__(self):
        target = self
        while target.__parent__ is not None:
            target = target.__parent__
        return target

    def transform(self):
        """make the little tuple instance get stronger(callable) ^_^"""
        func_finder = self.__root__
        for attr in self:
            func_finder = getattr(func_finder, attr)

        if isinstance(func_finder, self.__class__):
            return None
            # raise Exception("func called before resource setup."
            #                 " detail: %s" % str(self))

        return func_finder

    def __call__(self, *args, **kwargs):
        return self.transform()(*args, **kwargs)


def setup():
    global __setup__
    if __setup__:
        return

    finder = ResourceFinder()
    for path in finder.resource_path:
        install_resource(path)

    __setup__ = True
    resource.__finder__ = finder


def install_resource(rs_path):
    dotted_path = rs_path.path
    _resource = None
    endpoint = None

    # adapter, api
    if is_api(dotted_path) or is_adapter(dotted_path):
        return install_adapter(rs_path)

    # resource
    for p in dotted_path.split("."):
        if isinstance(_resource, ResourceShortcut):
            logger.debug("ignored: {}".format(dotted_path))
            rs_path.ignored()
            return
        _resource = getattr(_resource or resource, p)
        endpoint = p

    if _resource:
        try:
            resource_module = ResourceShortcut(".".join(_resource))
            logger.debug("success: {}".format(dotted_path))
            rs_path.loaded()
        except ResourceNotRegistered:
            logger.warning("failed: {}".format(dotted_path))
            rs_path.error()
            return

        # register shortcut
        shortcut = getattr(resource, endpoint)
        if isinstance(shortcut, ResourceShortcut):
            raise ResourceModuleConflict(
                "resources conflict:\n>>> {}\n<<< {}".format(shortcut._path, ".".join(_resource))
            )
        setattr(_resource.__parent__, endpoint, resource_module)
        setattr(resource, endpoint, resource_module)


def install_adapter(rs_path):
    dotted_path = rs_path.path
    adapter_cls = AdapterResourceShortcut
    # adapter 和 api 代码结构一致， 唯一区别是entry不同，adapter多了一层`adapter`目录
    if is_api(dotted_path):
        api_root = path_to_dotted(API_DIR)
        result = dotted_path[(len(API_DIR) + 1) :].split(".", 1)
        if len(result) == 2:
            rs, ada = result
        else:
            rs = result[0]
            ada = ""
        rs = "{}.{}".format(api_root, rs)
        adapter_cls = APIResourceShortcut
    else:
        rs, ada = [path.strip(".") for path in dotted_path.split("adapter")]

    try:
        default_adapter = adapter_cls(rs)
        defined_method = default_adapter.list_method()
    except ImportError as e:
        logger.warning("error: {}\n{}".format(dotted_path, e))
        rs_path.error()
        return

    if ada.startswith(settings.RUN_VER):
        platform_adapter = ResourceShortcut(dotted_path)
        # load method from platform adapter to default adapter
        for method in defined_method:
            if method in platform_adapter.list_method():
                default_adapter.reload_method(method, getattr(platform_adapter, method))

    root = adapter
    if is_api(dotted_path):
        root = api
    setattr(root, rs.split(".")[-1], default_adapter)
    logger.debug("success: {}".format(dotted_path))
    rs_path.loaded()


def is_api(dotted_path):
    return dotted_path.startswith(path_to_dotted(API_DIR))


def is_adapter(dotted_path):
    return "adapter" in dotted_path.split(".")


resource = ResourceManager()

adapter = ResourceManager()

api = ResourceManager()



# -*- coding: utf-8 -*-
"""
TencentBlueKing is pleased to support the open source community by making
蓝鲸智云 - Resource SDK (BlueKing - Resource SDK) available.
Copyright (C) 2023 THL A29 Limited,
a Tencent company. All rights reserved.
Licensed under the MIT License (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at https://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
either express or implied. See the License for the
specific language governing permissions and limitations under the License.
We undertake not to change the open source license (MIT license) applicable
to the current version of the project delivered to anyone in the future.
"""

from functools import wraps

from celery.result import AsyncResult
from celery.task import task
from django.utils.module_loading import import_string

from bk_resource.exceptions import CustomError
from bk_resource.utils.logger import logger
from bk_resource.utils.request import set_local_username


@task(bind=True)
def run_perform_request(self, resource_obj, username, request_data):
    """
    将resource作为异步任务执行
    :param self: 任务对象
    :param resource_obj: Resource实例
    :param username: 用户
    :param request_data: 请求数据
    :return: resource处理后的返回数据
    """
    if isinstance(resource_obj, str):
        resource_obj = import_string(resource_obj)()
    set_local_username(username)
    resource_obj._task_manager = self
    validated_request_data = resource_obj.validate_request_data(request_data)
    response_data = resource_obj.perform_request(validated_request_data)
    validated_response_data = resource_obj.validate_response_data(response_data)
    return validated_response_data


def _fetch_data_from_result(async_result):
    """
    从异步任务结果中提取步骤信息
    :param async_result: AsyncResult 对象
    :return: message, data: 步骤信息，步骤数据
    """
    info = async_result.info
    try:
        message = info.get("message")
        data = info.get("data")
    except Exception:
        message = None
        data = None
    return message, data


def query_task_result(task_id):
    """
    查询任务结果
    """
    result = AsyncResult(task_id)

    # 任务是否完成
    is_completed = False

    message, data = _fetch_data_from_result(result)

    if result.successful() or result.failed():
        is_completed = True
        try:
            # 任务执行完成，则读取结果数据
            message = None
            data = result.get()
        except CustomError as e:
            message = e.message
            data = e.data
        except Exception as e:
            logger.exception("Caught exception when running async resource task : %s" % e)
            message = "%s" % e
            data = None

    return {
        "task_id": task_id,
        "is_completed": is_completed,
        "state": result.state,
        "message": message,
        "data": data,
    }


def step(state=None, message=None, data=None):
    """
    步骤装饰器
    :param state: 当前状态，字符串
    :param message: 步骤信息
    :param data: 步骤数据
    """

    # 兼容不带括号的情况
    if callable(state):
        real_func = state
        real_state = None
    else:
        real_func = None
        real_state = state

    def decorate(func):
        @wraps(func)
        def wrapper(self, *args, **kwargs):
            # 不传递步骤名称，则默认使用函数名的全大写形式作为步骤名称
            _state = real_state or func.__name__.upper()
            # 更新任务状态
            self.update_state(state=_state, message=message, data=data)
            return func(self, *args, **kwargs)

        return wrapper

    if real_func:
        return decorate(real_func)

    return decorate



# -*- coding: utf-8 -*-
"""
TencentBlueKing is pleased to support the open source community by making
蓝鲸智云 - Resource SDK (BlueKing - Resource SDK) available.
Copyright (C) 2023 THL A29 Limited,
a Tencent company. All rights reserved.
Licensed under the MIT License (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at https://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
either express or implied. See the License for the
specific language governing permissions and limitations under the License.
We undertake not to change the open source license (MIT license) applicable
to the current version of the project delivered to anyone in the future.
"""

from functools import partial
from multiprocessing.pool import ThreadPool as _ThreadPool
from threading import Thread

from django import db
from django.utils import timezone, translation

from bk_resource.utils.local import local
from bk_resource.utils.logger import logger


class InheritParentThread(Thread):
    def __init__(self, *args, **kwargs):
        self.inherit_data = [item for item in local]
        self.timezone = timezone.get_current_timezone().zone
        self.language = translation.get_language()
        super(InheritParentThread, self).__init__(*args, **kwargs)

    def sync(self):
        for sync_item in self.inherit_data:
            setattr(local, sync_item[0], sync_item[1])
        timezone.activate(self.timezone)
        translation.activate(self.language)

    def unsync(self):
        # 新的线程会往local再写一些数据
        # 线程结束的时候，需要把所有线程相关的所有变量都清空
        for item in local:
            delattr(local, item[0])

        # db._connections 也是线程变量，所以在线程结束的时候需要主动的释放
        db.connections.close_all()

    def run(self):
        self.sync()
        try:
            super(InheritParentThread, self).run()
        except Exception as e:
            logger.exception(e)

        self.unsync()


def run_func_with_local(items, tz, lang, func, *args, **kwargs):
    """
    线程执行函数
    :param func: 待执行函数
    :param items: Thread Local Items
    :param tz: 时区
    :param lang: 语言
    :param args: 位置参数
    :param kwargs: 关键字参数
    :return: 函数返回值
    """
    # 同步local数据
    for item in items:
        setattr(local, item[0], item[1])

    # 设置时区及语言
    timezone.activate(tz)
    translation.activate(lang)

    try:
        data = func(*args, **kwargs)
    except Exception as e:
        raise e
    finally:
        # 关闭db连接
        db.connections.close_all()

        # 清理local数据
        for item in local:
            delattr(local, item[0])

    return data


class ThreadPool(_ThreadPool):
    """
    线程池
    """

    @staticmethod
    def get_func_with_local(func):
        tz = timezone.get_current_timezone().zone
        lang = translation.get_language()
        items = [item for item in local]
        return partial(run_func_with_local, items, tz, lang, func)

    def map_ignore_exception(self, func, iterable, return_exception=False):
        """
        忽略错误版的map
        """
        futures = []
        for params in iterable:
            if not isinstance(params, (tuple, list)):
                params = (params,)
            futures.append(self.apply_async(func, args=params))

        results = []
        for future in futures:
            try:
                results.append(future.get())
            except Exception as e:
                if return_exception:
                    results.append(e)
                logger.exception(e)

        return results

    def map_async(self, func, iterable, chunksize=None, callback=None, error_callback=None):
        return super(ThreadPool, self).map_async(
            self.get_func_with_local(func),
            iterable,
            chunksize=chunksize,
            callback=callback,
            error_callback=error_callback,
        )

    def apply_async(self, func, args=(), kwds=None, callback=None, error_callback=None):
        if kwds is None:
            kwds = {}
        return super(ThreadPool, self).apply_async(
            self.get_func_with_local(func), args=args, kwds=kwds, callback=callback, error_callback=error_callback
        )

    def imap(self, func, iterable, chunksize=1):
        return super(ThreadPool, self).imap(self.get_func_with_local(func), iterable, chunksize)

    def imap_unordered(self, func, iterable, chunksize=1):
        func = partial(run_func_with_local, func, local)
        return super(ThreadPool, self).imap_unordered(self.get_func_with_local(func), iterable, chunksize=chunksize)


if __name__ == "__main__":
    InheritParentThread().start()



# -*- coding: utf-8 -*-
"""
TencentBlueKing is pleased to support the open source community by making
蓝鲸智云 - Resource SDK (BlueKing - Resource SDK) available.
Copyright (C) 2023 THL A29 Limited,
a Tencent company. All rights reserved.
Licensed under the MIT License (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at https://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
either express or implied. See the License for the
specific language governing permissions and limitations under the License.
We undertake not to change the open source license (MIT license) applicable
to the current version of the project delivered to anyone in the future.
"""

import functools
import json
import time
import zlib

from django.core.cache import cache, caches
from django.utils.encoding import force_bytes
from django.utils.translation import gettext

from bk_resource.contrib.model import Empty
from bk_resource.settings import bk_resource_settings
from bk_resource.utils.common_utils import count_md5
from bk_resource.utils.local import local
from bk_resource.utils.logger import logger
from bk_resource.utils.request import get_request_username

try:
    mem_cache = caches["locmem"]
except Exception:
    mem_cache = cache


class UsingCache(object):
    min_length = 15
    preset = 6
    key_prefix = "web_cache"

    def __init__(
        self,
        cache_type,
        backend_cache_type=None,
        user_related=None,
        compress=True,
        is_cache_func=lambda res: True,
        func_key_generator=lambda func: "{}.{}".format(func.__module__, func.__name__),
    ):
        """
        :param cache_type: 缓存类型
        :param user_related: 是否与用户关联
        :param compress: 是否进行压缩
        :param is_cache_func: 缓存函数，当函数返回true时，则进行缓存
        :param func_key_generator: 函数标识key的生成逻辑
        """
        self.cache_type = cache_type
        self.backend_cache_type = backend_cache_type
        self.compress = compress
        self.is_cache_func = is_cache_func
        self.func_key_generator = func_key_generator
        # 先看用户是否提供了user_related参数
        # 若无，则查看cache_type是否提供了user_related参数
        # 若都没有定义，则user_related默认为True
        if user_related is not None:
            self.user_related = user_related
        elif getattr(cache_type, "user_related", None) is not None:
            self.user_related = self.cache_type.user_related
        else:
            self.user_related = True

        self.using_cache_type = self._get_using_cache_type()
        self.local_cache_enable = bool(bk_resource_settings.LOCAL_CACHE_ENABLE)

    def _get_username(self):
        username = "backend"
        if self.user_related:
            try:
                username = get_request_username()
            except Exception:
                username = "backend"
        return username

    def _get_using_cache_type(self):
        using_cache_type = self.cache_type
        if self._get_username() == "backend":
            using_cache_type = self.backend_cache_type or self.cache_type
        if using_cache_type:
            if not isinstance(using_cache_type, CacheTypeItem):
                raise TypeError("param 'cache_type' must be an" "instance of <utils.cache.CacheTypeItem>")
        return using_cache_type

    def _cache_key(self, task_definition, args, kwargs):
        # 新增根据用户openid设置缓存key
        if self.using_cache_type:
            return "{}:{}:{}:{},{}[{}]".format(
                self.key_prefix,
                self.using_cache_type.key,
                self.func_key_generator(task_definition),
                count_md5(args),
                count_md5(kwargs),
                self._get_username(),
            )
        return None

    def get_value(self, cache_key, default=None):
        """
        新增一级内存缓存（local）。在同一个请求(线程)中，优先使用内存缓存。
        一级缓存： local（web服务单次请求中生效）
        二级缓存： cache（60s生效）
        机制：
        local (miss), cache(miss): cache <- result
        local (miss), cache(hit): local <- result
        """
        if self.local_cache_enable:
            value = getattr(local, cache_key, None)
            if value:
                return json.loads(value)

        value = mem_cache.get(cache_key, default=None) or cache.get(cache_key, default=None)
        if value is None:
            return default
        if self.compress:
            try:
                value = zlib.decompress(value)
            except Exception:
                pass
            try:
                value = json.loads(force_bytes(value))
            except Exception:
                value = default
        if value and self.local_cache_enable:
            setattr(local, cache_key, json.dumps(value))
        return value

    def set_value(self, key, value, timeout=60):
        if self.compress:
            try:
                value = json.dumps(value)
            except Exception:
                logger.exception(gettext("[Cache]不支持序列化的类型: %s"), type(value))
                return False

            if len(value) > self.min_length:
                value = zlib.compress(value.encode())

        try:
            if mem_cache is not cache:
                mem_cache.set(key, value, 60)
            cache.set(key, value, timeout)
        except Exception as e:
            try:
                from blueapps.utils import get_request

                request_path = get_request().path
            except Exception:
                request_path = ""
            # 缓存出错不影响主流程
            logger.exception(
                gettext("存缓存[key:%s]时报错：%s\n value: %r\nurl: %s"),
                key,
                e,
                value,
                request_path,
            )

    def _cached(self, task_definition, args, kwargs):
        """
        【默认缓存模式】
        先检查是否缓存是否存在
        若存在，则直接返回缓存内容
        若不存在，则执行函数，并将结果回写到缓存中
        """
        cache_key = self._cache_key(task_definition, args, kwargs)
        if cache_key:
            return_value = self.get_value(cache_key)

            if return_value is None:
                return_value = self._refresh(task_definition, args, kwargs)
        else:
            return_value = self._cacheless(task_definition, args, kwargs)
        return return_value

    def _refresh(self, task_definition, args, kwargs):
        """
        【强制刷新模式】
        不使用缓存的数据，将函数执行返回结果回写缓存
        """
        cache_key = self._cache_key(task_definition, args, kwargs)

        return_value = self._cacheless(task_definition, args, kwargs)

        # 设置了缓存空数据
        # 或者不缓存空数据且数据为空时
        # 需要进行缓存
        if self.is_cache_func(return_value):
            self.set_value(cache_key, return_value, self.using_cache_type.timeout)

        return return_value

    def _cacheless(self, task_definition, args, kwargs):
        """
        【忽略缓存模式】
        忽略缓存机制，直接执行函数，返回结果不回写缓存
        """
        # 执行真实函数
        return task_definition(*args, **kwargs)

    def __call__(self, task_definition):
        @functools.wraps(task_definition)
        def cached_wrapper(*args, **kwargs):
            return_value = self._cached(task_definition, args, kwargs)
            return return_value

        @functools.wraps(task_definition)
        def refresh_wrapper(*args, **kwargs):
            return_value = self._refresh(task_definition, args, kwargs)
            return return_value

        @functools.wraps(task_definition)
        def cacheless_wrapper(*args, **kwargs):
            return_value = self._cacheless(task_definition, args, kwargs)
            return return_value

        # 为函数设置各种调用模式
        default_wrapper = cached_wrapper
        default_wrapper.cached = cached_wrapper
        default_wrapper.refresh = refresh_wrapper
        default_wrapper.cacheless = cacheless_wrapper

        return default_wrapper


using_cache = UsingCache


class CacheTypeItem(object):
    """
    缓存类型定义
    """

    def __init__(self, key, timeout, user_related=None, label=""):
        """
        :param key: 缓存名称
        :param timeout: 缓存超时，单位：s
        :param user_related: 是否用户相关
        :param label: 详细说明
        """
        self.key = key
        self.timeout = timeout
        self.label = label
        self.user_related = user_related

    def __call__(self, timeout):
        return CacheTypeItem(self.key, timeout, self.user_related, self.label)


class InstanceCache(object):
    _instance = Empty()

    @classmethod
    def instance(cls, *args, **kwargs):
        if isinstance(cls._instance, Empty):
            cls._instance = cls()
        return cls._instance

    def __init__(self):
        self.__cache = {}

    def clear(self):
        self.__cache = {}

    def set(self, key, value, seconds=0, use_round=False):
        """
        :param key:
        :param value:
        :param seconds:
        :param use_round: 时间是否需要向上取整，取整用于缓存时间同步
        :return:
        """
        if not seconds:
            timeout = 0
        else:
            if not use_round:
                timeout = time.time() + seconds
            else:
                timeout = (time.time() + seconds) // seconds * seconds
        self.__cache[key] = (value, timeout)

    def __get_raw(self, key):
        value = self.__cache.get(key)
        if not value:
            return None
        if value[1] and time.time() > value[1]:
            del self.__cache[key]
            return None
        return value

    def exists(self, key):
        value = self.__get_raw(key)
        return value is not None

    def get(self, key):
        value = self.__get_raw(key)
        return value and value[0]

    def delete(self, key):
        try:
            del self.__cache[key]
        except KeyError:
            pass



# -*- coding: utf-8 -*-
"""
TencentBlueKing is pleased to support the open source community by making
蓝鲸智云 - Resource SDK (BlueKing - Resource SDK) available.
Copyright (C) 2023 THL A29 Limited,
a Tencent company. All rights reserved.
Licensed under the MIT License (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at https://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
either express or implied. See the License for the
specific language governing permissions and limitations under the License.
We undertake not to change the open source license (MIT license) applicable
to the current version of the project delivered to anyone in the future.
"""

import abc

from bk_resource.base import Resource
from bk_resource.utils.cache import CacheTypeItem, using_cache


class CacheResource(Resource, metaclass=abc.ABCMeta):
    """
    支持缓存的resource
    """

    # 缓存类型
    cache_type = None
    # 后台缓存时间
    backend_cache_type = None
    # 缓存是否与用户关联，默认关联
    cache_user_related: bool = None
    # 是否使用压缩
    cache_compress = True

    def __init__(self, *args, **kwargs):
        # 若cache_type为None则视为关闭缓存功能
        if self._need_cache_wrap():
            self._wrap_request()
        super(CacheResource, self).__init__(*args, **kwargs)

    def _need_cache_wrap(self):
        need_cache = False
        if self.cache_type is not None:
            if not isinstance(self.cache_type, CacheTypeItem):
                raise TypeError("param 'cache_type' must be an" "instance of <utils.cache.CacheTypeItem>")
            need_cache = True
        if self.backend_cache_type is not None:
            if not isinstance(self.backend_cache_type, CacheTypeItem):
                raise TypeError("param 'cache_type' must be an" "instance of <utils.cache.CacheTypeItem>")
            need_cache = True
        return need_cache

    def _wrap_request(self):
        """
        将原有的request方法替换为支持缓存的request方法
        """

        def func_key_generator(resource):
            key = "{}.{}".format(
                resource.__self__.__class__.__module__,
                resource.__self__.__class__.__name__,
            )
            return key

        self.request = using_cache(
            cache_type=self.cache_type,
            backend_cache_type=self.backend_cache_type,
            user_related=self.cache_user_related,
            compress=self.cache_compress,
            is_cache_func=self.cache_write_trigger,
            func_key_generator=func_key_generator,
        )(self.request)

    def cache_write_trigger(self, res):
        """
        缓存写入触发条件
        """
        return True



# -*- coding: utf-8 -*-
"""
TencentBlueKing is pleased to support the open source community by making
蓝鲸智云 - Resource SDK (BlueKing - Resource SDK) available.
Copyright (C) 2023 THL A29 Limited,
a Tencent company. All rights reserved.
Licensed under the MIT License (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at https://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
either express or implied. See the License for the
specific language governing permissions and limitations under the License.
We undertake not to change the open source license (MIT license) applicable
to the current version of the project delivered to anyone in the future.
"""

import os

from django.apps import apps
from django.conf import settings
from django.contrib.staticfiles.finders import BaseFinder

from bk_resource.management.exceptions import ErrorSettingsWithResourceDirs
from bk_resource.settings import bk_resource_settings
from bk_resource.utils.text import path_to_dotted

API_DIR = bk_resource_settings.DEFAULT_API_DIR
RESOURCE_DIRS = bk_resource_settings.DEFAULT_RESOURCE_DIRS
if API_DIR not in RESOURCE_DIRS:
    RESOURCE_DIRS.append(API_DIR)


class ResourceFinder(BaseFinder):
    def __init__(self, app_names=None, *args, **kwargs):
        # Mapping of app names to resource modules
        self.resource_path = []
        app_path_directories = []
        app_configs = apps.get_app_configs()
        if app_names:
            app_names = set(app_names)
            app_configs = [ac for ac in app_configs if ac.name in app_names]

        for app_config in app_configs:
            self.resource_path += self.find(app_config.path, root_path=os.path.dirname(app_config.path))
            app_path_directories.append(app_config.path)

        for path in RESOURCE_DIRS:
            search_path = os.path.join(settings.BASE_DIR, path)
            if search_path in app_path_directories:
                continue

            self.resource_path += self.find(search_path, from_settings=True)

        self.resource_path.sort()
        self.found()

    def found(self):
        p_list = []
        for p in self.resource_path:
            p_list.append(ResourcePath(p))

        self.resource_path = p_list

    def list(self, ignore_patterns):
        """
        List all resource path.
        """
        for path in self.resource_path:
            yield path.path, path.status

    def find(self, path, root_path=None, from_settings=False):
        """
        Looks for resource module in the app directories.
        if recursive is True, recursive traversal directory
        """
        matches = set()
        if not path.startswith(settings.BASE_DIR):
            if from_settings:
                raise ErrorSettingsWithResourceDirs("RESOURCE_DIRS settings error")
            return []

        for root, dirs, files in sorted(os.walk(path)):
            if os.path.basename(root) == "resources":
                relative_path = os.path.relpath(os.path.dirname(root), root_path or settings.BASE_DIR)
                matches.add(path_to_dotted(relative_path))
                continue

            for file_path in files:
                file_name = os.path.basename(file_path)
                base_file_name, ext_name = os.path.splitext(file_name)
                if ext_name not in [".py", ".pyc", ".pyo"]:
                    continue

                if base_file_name in ["resources", "default"]:
                    relative_path = os.path.relpath(root, root_path or settings.BASE_DIR)
                    matches.add(path_to_dotted(relative_path))
                    continue

        return matches


class ResourceStatus(object):
    # 待加载
    unloaded = "unloaded"
    # 加载成功
    loaded = "loaded"
    # 忽略
    ignored = "ignored"
    # 加载错误
    error = "error"


class ResourcePath(object):
    """
    path = ResourcePath("api.xxx")
    path.loaded()
    path.ignored()
    path.error()
    """

    def __init__(self, path):
        status = ResourceStatus.unloaded
        path_info = path.split(":")
        if len(path_info) > 1:
            rspath, status = path_info[:2]
        else:
            rspath = path_info[0]

        self.path = rspath.strip()
        self.status = status.strip()

    def __getattr__(self, item):
        status = getattr(ResourceStatus, item, None)
        if status:
            return status_setter(status)(lambda: status, self)

    def __repr__(self):
        return "{}: {}".format(self.path, self.status)


def status_setter(status):
    def setter(func, path):
        path.status = status
        return func

    return setter



# -*- coding: utf-8 -*-
"""
TencentBlueKing is pleased to support the open source community by making
蓝鲸智云 - Resource SDK (BlueKing - Resource SDK) available.
Copyright (C) 2023 THL A29 Limited,
a Tencent company. All rights reserved.
Licensed under the MIT License (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at https://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
either express or implied. See the License for the
specific language governing permissions and limitations under the License.
We undertake not to change the open source license (MIT license) applicable
to the current version of the project delivered to anyone in the future.
"""

from drf_yasg import openapi
from drf_yasg.generators import OpenAPISchemaGenerator


class BKResourceOpenAPISchemaGenerator(OpenAPISchemaGenerator):
    def get_path_item(self, path, view_cls, operations):
        """Get a :class:`.PathItem` object that describes the parameters and operations related to a single path in the
        API.

        :param str path: the path
        :param type view_cls: the view that was bound to this path in urlpatterns
        :param dict[str,openapi.Operation] operations: operations defined on this path, keyed by lowercase HTTP method
        :rtype: openapi.PathItem
        """
        path_parameters = self.get_path_parameters(path, view_cls)

        # 移除get参数中与path_parameters重复的项
        if operations.get("get"):
            paths_params = [param["name"] for param in path_parameters]
            query_params = []
            for path_param in operations["get"]["parameters"]:
                if path_param.name not in paths_params:
                    query_params.append(path_param)
            operations["get"]["parameters"] = query_params
        return openapi.PathItem(parameters=path_parameters, **operations)



# -*- coding: utf-8 -*-
"""
TencentBlueKing is pleased to support the open source community by making
蓝鲸智云 - Resource SDK (BlueKing - Resource SDK) available.
Copyright (C) 2023 THL A29 Limited,
a Tencent company. All rights reserved.
Licensed under the MIT License (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at https://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
either express or implied. See the License for the
specific language governing permissions and limitations under the License.
We undertake not to change the open source license (MIT license) applicable
to the current version of the project delivered to anyone in the future.
"""

import abc
import json

import requests
from django.conf import settings

from bk_resource.contrib.api import APIResource
from bk_resource.exceptions import IAMNoPermission, PlatformAuthParamsNotExist
from bk_resource.settings import bk_resource_settings
from bk_resource.utils.common_utils import is_backend


class BkApiResource(APIResource, abc.ABC):
    method = "GET"
    bkapi_header_authorization = False
    platform_authorization = False

    def add_platform_auth_params(self, params, force_platform_auth=False):
        """
        在以下情况不使用
        1. 未开启了平台鉴权
        2. API 未启用平台鉴权且不是后台任务
        """
        enable_platform_auth = bk_resource_settings.PLATFORM_AUTH_ENABLED
        if not enable_platform_auth or (not self.platform_authorization and not force_platform_auth):
            return params
        # 更新平台鉴权信息
        for _key in self.oath_cookies_params.keys():
            params.pop(_key, None)
        auth_token = bk_resource_settings.PLATFORM_AUTH_ACCESS_TOKEN
        auth_username = bk_resource_settings.PLATFORM_AUTH_ACCESS_USERNAME
        if not auth_token and not auth_username:
            raise PlatformAuthParamsNotExist()
        if auth_token:
            params["access_token"] = auth_token
        if auth_username:
            params["bk_username"] = auth_username
        return params

    def add_esb_info_before_request(self, params):
        """
        通过 params 参数控制是否检查 request
        @param {Boolean} [params.no_request] 是否需要带上 request 标识
        """
        params["bk_app_code"] = settings.APP_CODE
        params["bk_app_secret"] = settings.SECRET_KEY

        # 后台程序或非request请求直接返回
        if is_backend() or params.get("_is_backend", False):
            params.pop("_request", None)
            params = self.add_platform_auth_params(params, force_platform_auth=True)
            return params

        # 前端应用, _request，用于并发请求的场景
        from blueapps.utils.request_provider import get_local_request

        _request = params.get("_request")
        req = _request or get_local_request()
        auth_info = self.build_auth_args(req)
        params.update(auth_info)
        params["bk_username"] = getattr(req.user, "bk_username", None) or req.user.username
        params.pop("_request", None)

        # 平台鉴权兼容
        params = self.add_platform_auth_params(params)

        return params

    def build_auth_args(self, request):
        """组装认证信息"""
        auth_args = {}
        if request is None:
            return auth_args

        for key, value in self.oath_cookies_params.items():
            if value in request.COOKIES:
                auth_args.update({key: request.COOKIES[value]})
        return auth_args

    @property
    def oath_cookies_params(self):
        return getattr(settings, "OAUTH_COOKIES_PARAMS", {"bk_token": "bk_token"})

    def build_request_data(self, validated_request_data):
        return self.add_esb_info_before_request(validated_request_data)

    def build_header(self, validated_request_data):
        # 初始化
        headers = {}
        # 透传 Cookie
        from blueapps.utils.request_provider import get_local_request

        request = get_local_request()
        if request is not None:
            headers["cookie"] = "; ".join(
                [
                    f"{key}={val}"
                    for key, val in request.COOKIES.items()
                    if key in bk_resource_settings.REQUEST_BKAPI_COOKIE_FIELDS
                ]
            )
        # 鉴权参数
        if not self.bkapi_header_authorization:
            return headers
        return {
            **headers,
            "x-bkapi-authorization": json.dumps(
                self.add_platform_auth_params(
                    {
                        "bk_app_code": settings.APP_CODE,
                        "bk_app_secret": settings.SECRET_KEY,
                    }
                )
            ),
        }

    def parse_response(self, response: requests.Response):
        """
        兼容IAM无权限处理
        """

        try:
            result_json = response.json()
        except Exception:
            result_json = []

        if isinstance(result_json, dict) and str(result_json.get("code")) == IAMNoPermission().code:
            data = result_json.get("data", {})
            if result_json.get("permission"):
                data["permission"] = result_json["permission"]
            raise IAMNoPermission(data=json.dumps(data))

        return super().parse_response(response)
